"""This is invoked in a subprocess to call the build backend hooks.

It expects:
- Command line args: hook_name, control_dir
- Environment variables:
      PEP517_BUILD_BACKEND=entry.point:spec
      PEP517_BACKEND_PATH=paths (separated with os.pathsep)
- control_dir/input.json:
  - {"kwargs": {...}}

Results:
- control_dir/output.json
  - {"return_val": ...}
"""
import json
import os
import os.path
import re
import shutil
import sys
import traceback
from glob import glob
from importlib import import_module
from os.path import join as pjoin

# This file is run as a script, and `import wrappers` is not zip-safe, so we
# include write_json() and read_json() from wrappers.py.


def write_json(obj, path, **kwargs):
    with open(path, "w", encoding="utf-8") as f:
        json.dump(obj, f, **kwargs)


def read_json(path):
    with open(path, encoding="utf-8") as f:
        return json.load(f)


class BackendUnavailable(Exception):
    """Raised if we cannot import the backend"""

    def __init__(self, traceback):
        self.traceback = traceback


class BackendInvalid(Exception):
    """Raised if the backend is invalid"""

    def __init__(self, message):
        self.message = message


class HookMissing(Exception):
    """Raised if a hook is missing and we are not executing the fallback"""

    def __init__(self, hook_name=None):
        super().__init__(hook_name)
        self.hook_name = hook_name


def contained_in(filename, directory):
    """Test if a file is located within the given directory."""
    filename = os.path.normcase(os.path.abspath(filename))
    directory = os.path.normcase(os.path.abspath(directory))
    return os.path.commonprefix([filename, directory]) == directory


def _build_backend():
    """Find and load the build backend"""
    # Add in-tree backend directories to the front of sys.path.
    backend_path = os.environ.get("PEP517_BACKEND_PATH")
    if backend_path:
        extra_pathitems = backend_path.split(os.pathsep)
        sys.path[:0] = extra_pathitems

    ep = os.environ["PEP517_BUILD_BACKEND"]
    mod_path, _, obj_path = ep.partition(":")
    try:
        obj = import_module(mod_path)
    except ImportError:
        raise BackendUnavailable(traceback.format_exc())

    if backend_path:
        if not any(contained_in(obj.__file__, path) for path in extra_pathitems):
            raise BackendInvalid("Backend was not loaded from backend-path")

    if obj_path:
        for path_part in obj_path.split("."):
            obj = getattr(obj, path_part)
    return obj


def _supported_features():
    """Return the list of options features supported by the backend.

    Returns a list of strings.
    The only possible value is 'build_editable'.
    """
    backend = _build_backend()
    features = []
    if hasattr(backend, "build_editable"):
        features.append("build_editable")
    return features


def get_requires_for_build_wheel(config_settings):
    """Invoke the optional get_requires_for_build_wheel hook

    Returns [] if the hook is not defined.
    """
    backend = _build_backend()
    try:
        hook = backend.get_requires_for_build_wheel
    except AttributeError:
        return []
    else:
        return hook(config_settings)


def get_requires_for_build_editable(config_settings):
    """Invoke the optional get_requires_for_build_editable hook

    Returns [] if the hook is not defined.
    """
    backend = _build_backend()
    try:
        hook = backend.get_requires_for_build_editable
    except AttributeError:
        return []
    else:
        return hook(config_settings)


def prepare_metadata_for_build_wheel(
    metadata_directory, config_settings, _allow_fallback
):
    """Invoke optional prepare_metadata_for_build_wheel

    Implements a fallback by building a wheel if the hook isn't defined,
    unless _allow_fallback is False in which case HookMissing is raised.
    """
    backend = _build_backend()
    try:
        hook = backend.prepare_metadata_for_build_wheel
    except AttributeError:
        if not _allow_fallback:
            raise HookMissing()
        whl_basename = backend.build_wheel(metadata_directory, config_settings)
        return _get_wheel_metadata_from_wheel(
            whl_basename, metadata_directory, config_settings
        )
    else:
        return hook(metadata_directory, config_settings)


def prepare_metadata_for_build_editable(
    metadata_directory, config_settings, _allow_fallback
):
    """Invoke optional prepare_metadata_for_build_editable

    Implements a fallback by building an editable wheel if the hook isn't
    defined, unless _allow_fallback is False in which case HookMissing is
    raised.
    """
    backend = _build_backend()
    try:
        hook = backend.prepare_metadata_for_build_editable
    except AttributeError:
        if not _allow_fallback:
            raise HookMissing()
        try:
            build_hook = backend.build_editable
        except AttributeError:
            raise HookMissing(hook_name="build_editable")
        else:
            whl_basename = build_hook(metadata_directory, config_settings)
            return _get_wheel_metadata_from_wheel(
                whl_basename, metadata_directory, config_settings
            )
    else:
        return hook(metadata_directory, config_settings)


WHEEL_BUILT_MARKER = "PEP517_ALREADY_BUILT_WHEEL"


def _dist_info_files(whl_zip):
    """Identify the .dist-info folder inside a wheel ZipFile."""
    res = []
    for path in whl_zip.namelist():
        m = re.match(r"[^/\\]+-[^/\\]+\.dist-info/", path)
        if m:
            res.append(path)
    if res:
        return res
    raise Exception("No .dist-info folder found in wheel")


def _get_wheel_metadata_from_wheel(whl_basename, metadata_directory, config_settings):
    """Extract the metadata from a wheel.

    Fallback for when the build backend does not
    define the 'get_wheel_metadata' hook.
    """
    from zipfile import ZipFile

    with open(os.path.join(metadata_directory, WHEEL_BUILT_MARKER), "wb"):
        pass  # Touch marker file

    whl_file = os.path.join(metadata_directory, whl_basename)
    with ZipFile(whl_file) as zipf:
        dist_info = _dist_info_files(zipf)
        zipf.extractall(path=metadata_directory, members=dist_info)
    return dist_info[0].split("/")[0]


def _find_already_built_wheel(metadata_directory):
    """Check for a wheel already built during the get_wheel_metadata hook."""
    if not metadata_directory:
        return None
    metadata_parent = os.path.dirname(metadata_directory)
    if not os.path.isfile(pjoin(metadata_parent, WHEEL_BUILT_MARKER)):
        return None

    whl_files = glob(os.path.join(metadata_parent, "*.whl"))
    if not whl_files:
        print("Found wheel built marker, but no .whl files")
        return None
    if len(whl_files) > 1:
        print(
            "Found multiple .whl files; unspecified behaviour. "
            "Will call build_wheel."
        )
        return None

    # Exactly one .whl file
    return whl_files[0]


def build_wheel(wheel_directory, config_settings, metadata_directory=None):
    """Invoke the mandatory build_wheel hook.

    If a wheel was already built in the
    prepare_metadata_for_build_wheel fallback, this
    will copy it rather than rebuilding the wheel.
    """
    prebuilt_whl = _find_already_built_wheel(metadata_directory)
    if prebuilt_whl:
        shutil.copy2(prebuilt_whl, wheel_directory)
        return os.path.basename(prebuilt_whl)

    return _build_backend().build_wheel(
        wheel_directory, config_settings, metadata_directory
    )


def build_editable(wheel_directory, config_settings, metadata_directory=None):
    """Invoke the optional build_editable hook.

    If a wheel was already built in the
    prepare_metadata_for_build_editable fallback, this
    will copy it rather than rebuilding the wheel.
    """
    backend = _build_backend()
    try:
        hook = backend.build_editable
    except AttributeError:
        raise HookMissing()
    else:
        prebuilt_whl = _find_already_built_wheel(metadata_directory)
        if prebuilt_whl:
            shutil.copy2(prebuilt_whl, wheel_directory)
            return os.path.basename(prebuilt_whl)

        return hook(wheel_directory, config_settings, metadata_directory)


def get_requires_for_build_sdist(config_settings):
    """Invoke the optional get_requires_for_build_wheel hook

    Returns [] if the hook is not defined.
    """
    backend = _build_backend()
    try:
        hook = backend.get_requires_for_build_sdist
    except AttributeError:
        return []
    else:
        return hook(config_settings)


class _DummyException(Exception):
    """Nothing should ever raise this exception"""


class GotUnsupportedOperation(Exception):
    """For internal use when backend raises UnsupportedOperation"""

    def __init__(self, traceback):
        self.traceback = traceback


def build_sdist(sdist_directory, config_settings):
    """Invoke the mandatory build_sdist hook."""
    backend = _build_backend()
    try:
        return backend.build_sdist(sdist_directory, config_settings)
    except getattr(backend, "UnsupportedOperation", _DummyException):
        raise GotUnsupportedOperation(traceback.format_exc())


HOOK_NAMES = {
    "get_requires_for_build_wheel",
    "prepare_metadata_for_build_wheel",
    "build_wheel",
    "get_requires_for_build_editable",
    "prepare_metadata_for_build_editable",
    "build_editable",
    "get_requires_for_build_sdist",
    "build_sdist",
    "_supported_features",
}


def main():
    if len(sys.argv) < 3:
        sys.exit("Needs args: hook_name, control_dir")
    hook_name = sys.argv[1]
    control_dir = sys.argv[2]
    if hook_name not in HOOK_NAMES:
        sys.exit("Unknown hook: %s" % hook_name)
    hook = globals()[hook_name]

    hook_input = read_json(pjoin(control_dir, "input.json"))

    json_out = {"unsupported": False, "return_val": None}
    try:
        json_out["return_val"] = hook(**hook_input["kwargs"])
    except BackendUnavailable as e:
        json_out["no_backend"] = True
        json_out["traceback"] = e.traceback
    except BackendInvalid as e:
        json_out["backend_invalid"] = True
        json_out["backend_error"] = e.message
    except GotUnsupportedOperation as e:
        json_out["unsupported"] = True
        json_out["traceback"] = e.traceback
    except HookMissing as e:
        json_out["hook_missing"] = True
        json_out["missing_hook_name"] = e.hook_name or hook_name

    write_json(json_out, pjoin(control_dir, "output.json"), indent=2)


if __name__ == "__main__":
    main()



import io
from typing import IO, TYPE_CHECKING, Any, List

from .ansi import AnsiDecoder
from .text import Text

if TYPE_CHECKING:
    from .console import Console


class FileProxy(io.TextIOBase):
    """Wraps a file (e.g. sys.stdout) and redirects writes to a console."""

    def __init__(self, console: "Console", file: IO[str]) -> None:
        self.__console = console
        self.__file = file
        self.__buffer: List[str] = []
        self.__ansi_decoder = AnsiDecoder()

    @property
    def rich_proxied_file(self) -> IO[str]:
        """Get proxied file."""
        return self.__file

    def __getattr__(self, name: str) -> Any:
        return getattr(self.__file, name)

    def write(self, text: str) -> int:
        if not isinstance(text, str):
            raise TypeError(f"write() argument must be str, not {type(text).__name__}")
        buffer = self.__buffer
        lines: List[str] = []
        while text:
            line, new_line, text = text.partition("\n")
            if new_line:
                lines.append("".join(buffer) + line)
                del buffer[:]
            else:
                buffer.append(line)
                break
        if lines:
            console = self.__console
            with console:
                output = Text("\n").join(
                    self.__ansi_decoder.decode_line(line) for line in lines
                )
                console.print(output)
        return len(text)

    def flush(self) -> None:
        output = "".join(self.__buffer)
        if output:
            self.__console.print(output)
        del self.__buffer[:]



# -*- coding: utf-8 -*-
#
# Copyright (C) 2013-2015 Vinay Sajip.
# Licensed to the Python Software Foundation under a contributor agreement.
# See LICENSE.txt and CONTRIBUTORS.txt.
#
from io import BytesIO
import logging
import os
import re
import struct
import sys
import time
from zipfile import ZipInfo

from .compat import sysconfig, detect_encoding, ZipFile
from .resources import finder
from .util import (
    FileOperator,
    get_export_entry,
    convert_path,
    get_executable,
    get_platform,
    in_venv,
)

logger = logging.getLogger(__name__)

_DEFAULT_MANIFEST = """
<?xml version="1.0" encoding="UTF-8" standalone="yes"?>
<assembly xmlns="urn:schemas-microsoft-com:asm.v1" manifestVersion="1.0">
 <assemblyIdentity version="1.0.0.0"
 processorArchitecture="X86"
 name="%s"
 type="win32"/>

 <!-- Identify the application security requirements. -->
 <trustInfo xmlns="urn:schemas-microsoft-com:asm.v3">
 <security>
 <requestedPrivileges>
 <requestedExecutionLevel level="asInvoker" uiAccess="false"/>
 </requestedPrivileges>
 </security>
 </trustInfo>
</assembly>""".strip()

# check if Python is called on the first line with this expression
FIRST_LINE_RE = re.compile(b"^#!.*pythonw?[0-9.]*([ \t].*)?$")
SCRIPT_TEMPLATE = r"""# -*- coding: utf-8 -*-
import re
import sys
from %(module)s import %(import_name)s
if __name__ == '__main__':
    sys.argv[0] = re.sub(r'(-script\.pyw|\.exe)?$', '', sys.argv[0])
    sys.exit(%(func)s())
"""


def enquote_executable(executable):
    if " " in executable:
        # make sure we quote only the executable in case of env
        # for example /usr/bin/env "/dir with spaces/bin/jython"
        # instead of "/usr/bin/env /dir with spaces/bin/jython"
        # otherwise whole
        if executable.startswith("/usr/bin/env "):
            env, _executable = executable.split(" ", 1)
            if " " in _executable and not _executable.startswith('"'):
                executable = '%s "%s"' % (env, _executable)
        else:
            if not executable.startswith('"'):
                executable = '"%s"' % executable
    return executable


# Keep the old name around (for now), as there is at least one project using it!
_enquote_executable = enquote_executable


class ScriptMaker(object):
    """
    A class to copy or create scripts from source scripts or callable
    specifications.
    """

    script_template = SCRIPT_TEMPLATE

    executable = None  # for shebangs

    def __init__(
        self, source_dir, target_dir, add_launchers=True, dry_run=False, fileop=None
    ):
        self.source_dir = source_dir
        self.target_dir = target_dir
        self.add_launchers = add_launchers
        self.force = False
        self.clobber = False
        # It only makes sense to set mode bits on POSIX.
        self.set_mode = (os.name == "posix") or (
            os.name == "java" and os._name == "posix"
        )
        self.variants = set(("", "X.Y"))
        self._fileop = fileop or FileOperator(dry_run)

        self._is_nt = os.name == "nt" or (os.name == "java" and os._name == "nt")
        self.version_info = sys.version_info

    def _get_alternate_executable(self, executable, options):
        if options.get("gui", False) and self._is_nt:  # pragma: no cover
            dn, fn = os.path.split(executable)
            fn = fn.replace("python", "pythonw")
            executable = os.path.join(dn, fn)
        return executable

    if sys.platform.startswith("java"):  # pragma: no cover

        def _is_shell(self, executable):
            """
            Determine if the specified executable is a script
            (contains a #! line)
            """
            try:
                with open(executable) as fp:
                    return fp.read(2) == "#!"
            except (OSError, IOError):
                logger.warning("Failed to open %s", executable)
                return False

        def _fix_jython_executable(self, executable):
            if self._is_shell(executable):
                # Workaround for Jython is not needed on Linux systems.
                import java

                if java.lang.System.getProperty("os.name") == "Linux":
                    return executable
            elif executable.lower().endswith("jython.exe"):
                # Use wrapper exe for Jython on Windows
                return executable
            return "/usr/bin/env %s" % executable

    def _build_shebang(self, executable, post_interp):
        """
        Build a shebang line. In the simple case (on Windows, or a shebang line
        which is not too long or contains spaces) use a simple formulation for
        the shebang. Otherwise, use /bin/sh as the executable, with a contrived
        shebang which allows the script to run either under Python or sh, using
        suitable quoting. Thanks to Harald Nordgren for his input.

        See also: http://www.in-ulm.de/~mascheck/various/shebang/#length
                  https://hg.mozilla.org/mozilla-central/file/tip/mach
        """
        if os.name != "posix":
            simple_shebang = True
        else:
            # Add 3 for '#!' prefix and newline suffix.
            shebang_length = len(executable) + len(post_interp) + 3
            if sys.platform == "darwin":
                max_shebang_length = 512
            else:
                max_shebang_length = 127
            simple_shebang = (b" " not in executable) and (
                shebang_length <= max_shebang_length
            )

        if simple_shebang:
            result = b"#!" + executable + post_interp + b"\n"
        else:
            result = b"#!/bin/sh\n"
            result += b"'''exec' " + executable + post_interp + b' "$0" "$@"\n'
            result += b"' '''"
        return result

    def _get_shebang(self, encoding, post_interp=b"", options=None):
        enquote = True
        if self.executable:
            executable = self.executable
            enquote = False  # assume this will be taken care of
        elif not sysconfig.is_python_build():
            executable = get_executable()
        elif in_venv():  # pragma: no cover
            executable = os.path.join(
                sysconfig.get_path("scripts"),
                "python%s" % sysconfig.get_config_var("EXE"),
            )
        else:  # pragma: no cover
            executable = os.path.join(
                sysconfig.get_config_var("BINDIR"),
                "python%s%s"
                % (
                    sysconfig.get_config_var("VERSION"),
                    sysconfig.get_config_var("EXE"),
                ),
            )
            if not os.path.isfile(executable):
                # for Python builds from source on Windows, no Python executables with
                # a version suffix are created, so we use python.exe
                executable = os.path.join(
                    sysconfig.get_config_var("BINDIR"),
                    "python%s" % (sysconfig.get_config_var("EXE")),
                )
        if options:
            executable = self._get_alternate_executable(executable, options)

        if sys.platform.startswith("java"):  # pragma: no cover
            executable = self._fix_jython_executable(executable)

        # Normalise case for Windows - COMMENTED OUT
        # executable = os.path.normcase(executable)
        # N.B. The normalising operation above has been commented out: See
        # issue #124. Although paths in Windows are generally case-insensitive,
        # they aren't always. For example, a path containing a ẞ (which is a
        # LATIN CAPITAL LETTER SHARP S - U+1E9E) is normcased to ß (which is a
        # LATIN SMALL LETTER SHARP S' - U+00DF). The two are not considered by
        # Windows as equivalent in path names.

        # If the user didn't specify an executable, it may be necessary to
        # cater for executable paths with spaces (not uncommon on Windows)
        if enquote:
            executable = enquote_executable(executable)
        # Issue #51: don't use fsencode, since we later try to
        # check that the shebang is decodable using utf-8.
        executable = executable.encode("utf-8")
        # in case of IronPython, play safe and enable frames support
        if (
            sys.platform == "cli"
            and "-X:Frames" not in post_interp
            and "-X:FullFrames" not in post_interp
        ):  # pragma: no cover
            post_interp += b" -X:Frames"
        shebang = self._build_shebang(executable, post_interp)
        # Python parser starts to read a script using UTF-8 until
        # it gets a #coding:xxx cookie. The shebang has to be the
        # first line of a file, the #coding:xxx cookie cannot be
        # written before. So the shebang has to be decodable from
        # UTF-8.
        try:
            shebang.decode("utf-8")
        except UnicodeDecodeError:  # pragma: no cover
            raise ValueError("The shebang (%r) is not decodable from utf-8" % shebang)
        # If the script is encoded to a custom encoding (use a
        # #coding:xxx cookie), the shebang has to be decodable from
        # the script encoding too.
        if encoding != "utf-8":
            try:
                shebang.decode(encoding)
            except UnicodeDecodeError:  # pragma: no cover
                raise ValueError(
                    "The shebang (%r) is not decodable "
                    "from the script encoding (%r)" % (shebang, encoding)
                )
        return shebang

    def _get_script_text(self, entry):
        return self.script_template % dict(
            module=entry.prefix,
            import_name=entry.suffix.split(".")[0],
            func=entry.suffix,
        )

    manifest = _DEFAULT_MANIFEST

    def get_manifest(self, exename):
        base = os.path.basename(exename)
        return self.manifest % base

    def _write_script(self, names, shebang, script_bytes, filenames, ext):
        use_launcher = self.add_launchers and self._is_nt
        linesep = os.linesep.encode("utf-8")
        if not shebang.endswith(linesep):
            shebang += linesep
        if not use_launcher:
            script_bytes = shebang + script_bytes
        else:  # pragma: no cover
            if ext == "py":
                launcher = self._get_launcher("t")
            else:
                launcher = self._get_launcher("w")
            stream = BytesIO()
            with ZipFile(stream, "w") as zf:
                source_date_epoch = os.environ.get("SOURCE_DATE_EPOCH")
                if source_date_epoch:
                    date_time = time.gmtime(int(source_date_epoch))[:6]
                    zinfo = ZipInfo(filename="__main__.py", date_time=date_time)
                    zf.writestr(zinfo, script_bytes)
                else:
                    zf.writestr("__main__.py", script_bytes)
            zip_data = stream.getvalue()
            script_bytes = launcher + shebang + zip_data
        for name in names:
            outname = os.path.join(self.target_dir, name)
            if use_launcher:  # pragma: no cover
                n, e = os.path.splitext(outname)
                if e.startswith(".py"):
                    outname = n
                outname = "%s.exe" % outname
                try:
                    self._fileop.write_binary_file(outname, script_bytes)
                except Exception:
                    # Failed writing an executable - it might be in use.
                    logger.warning(
                        "Failed to write executable - trying to " "use .deleteme logic"
                    )
                    dfname = "%s.deleteme" % outname
                    if os.path.exists(dfname):
                        os.remove(dfname)  # Not allowed to fail here
                    os.rename(outname, dfname)  # nor here
                    self._fileop.write_binary_file(outname, script_bytes)
                    logger.debug("Able to replace executable using " ".deleteme logic")
                    try:
                        os.remove(dfname)
                    except Exception:
                        pass  # still in use - ignore error
            else:
                if self._is_nt and not outname.endswith("." + ext):  # pragma: no cover
                    outname = "%s.%s" % (outname, ext)
                if os.path.exists(outname) and not self.clobber:
                    logger.warning("Skipping existing file %s", outname)
                    continue
                self._fileop.write_binary_file(outname, script_bytes)
                if self.set_mode:
                    self._fileop.set_executable_mode([outname])
            filenames.append(outname)

    variant_separator = "-"

    def get_script_filenames(self, name):
        result = set()
        if "" in self.variants:
            result.add(name)
        if "X" in self.variants:
            result.add("%s%s" % (name, self.version_info[0]))
        if "X.Y" in self.variants:
            result.add(
                "%s%s%s.%s"
                % (
                    name,
                    self.variant_separator,
                    self.version_info[0],
                    self.version_info[1],
                )
            )
        return result

    def _make_script(self, entry, filenames, options=None):
        post_interp = b""
        if options:
            args = options.get("interpreter_args", [])
            if args:
                args = " %s" % " ".join(args)
                post_interp = args.encode("utf-8")
        shebang = self._get_shebang("utf-8", post_interp, options=options)
        script = self._get_script_text(entry).encode("utf-8")
        scriptnames = self.get_script_filenames(entry.name)
        if options and options.get("gui", False):
            ext = "pyw"
        else:
            ext = "py"
        self._write_script(scriptnames, shebang, script, filenames, ext)

    def _copy_script(self, script, filenames):
        adjust = False
        script = os.path.join(self.source_dir, convert_path(script))
        outname = os.path.join(self.target_dir, os.path.basename(script))
        if not self.force and not self._fileop.newer(script, outname):
            logger.debug("not copying %s (up-to-date)", script)
            return

        # Always open the file, but ignore failures in dry-run mode --
        # that way, we'll get accurate feedback if we can read the
        # script.
        try:
            f = open(script, "rb")
        except IOError:  # pragma: no cover
            if not self.dry_run:
                raise
            f = None
        else:
            first_line = f.readline()
            if not first_line:  # pragma: no cover
                logger.warning("%s is an empty file (skipping)", script)
                return

            match = FIRST_LINE_RE.match(first_line.replace(b"\r\n", b"\n"))
            if match:
                adjust = True
                post_interp = match.group(1) or b""

        if not adjust:
            if f:
                f.close()
            self._fileop.copy_file(script, outname)
            if self.set_mode:
                self._fileop.set_executable_mode([outname])
            filenames.append(outname)
        else:
            logger.info("copying and adjusting %s -> %s", script, self.target_dir)
            if not self._fileop.dry_run:
                encoding, lines = detect_encoding(f.readline)
                f.seek(0)
                shebang = self._get_shebang(encoding, post_interp)
                if b"pythonw" in first_line:  # pragma: no cover
                    ext = "pyw"
                else:
                    ext = "py"
                n = os.path.basename(outname)
                self._write_script([n], shebang, f.read(), filenames, ext)
            if f:
                f.close()

    @property
    def dry_run(self):
        return self._fileop.dry_run

    @dry_run.setter
    def dry_run(self, value):
        self._fileop.dry_run = value

    if os.name == "nt" or (os.name == "java" and os._name == "nt"):  # pragma: no cover
        # Executable launcher support.
        # Launchers are from https://bitbucket.org/vinay.sajip/simple_launcher/

        def _get_launcher(self, kind):
            if struct.calcsize("P") == 8:  # 64-bit
                bits = "64"
            else:
                bits = "32"
            platform_suffix = "-arm" if get_platform() == "win-arm64" else ""
            name = "%s%s%s.exe" % (kind, bits, platform_suffix)
            # Issue 31: don't hardcode an absolute package name, but
            # determine it relative to the current package
            distlib_package = __name__.rsplit(".", 1)[0]
            resource = finder(distlib_package).find(name)
            if not resource:
                msg = "Unable to find resource %s in package %s" % (
                    name,
                    distlib_package,
                )
                raise ValueError(msg)
            return resource.bytes

    # Public API follows

    def make(self, specification, options=None):
        """
        Make a script.

        :param specification: The specification, which is either a valid export
                              entry specification (to make a script from a
                              callable) or a filename (to make a script by
                              copying from a source location).
        :param options: A dictionary of options controlling script generation.
        :return: A list of all absolute pathnames written to.
        """
        filenames = []
        entry = get_export_entry(specification)
        if entry is None:
            self._copy_script(specification, filenames)
        else:
            self._make_script(entry, filenames, options=options)
        return filenames

    def make_multiple(self, specifications, options=None):
        """
        Take a list of specifications and make scripts from them,
        :param specifications: A list of specifications.
        :return: A list of all absolute pathnames written to,
        """
        filenames = []
        for specification in specifications:
            filenames.extend(self.make(specification, options))
        return filenames



import hashlib
import json
import os

from spinsrv import spin
from spinsrv import spinpy as sp

if "SPINPY_TEST_PUBLIC" not in os.environ:
    raise ValueError("must specify SPINPY_TEST_PUBLIC environment variable")
if "SPINPY_TEST_PRIVATE" not in os.environ:
    raise ValueError("must specify SPINPY_TEST_PRIVATE environment variable")
if "SPINPY_TEST_CITIZEN" not in os.environ:
    raise ValueError("must specify SPINPY_TEST_CITIZEN environment variable")
SPINPY_TEST_PUBLIC = os.environ["SPINPY_TEST_PUBLIC"]
SPINPY_TEST_PRIVATE = os.environ["SPINPY_TEST_PRIVATE"]
SPINPY_TEST_CITIZEN = os.environ["SPINPY_TEST_CITIZEN"]

print("THIS IS SMOKE TEST 1: IT TESTS THE CLIENTS IN src/spin/spinpy.py")

kc = sp.KeyServerHTTPClient()
resp = kc.which(
    spin.KeyWhichRequest(
        public=SPINPY_TEST_PUBLIC,
        private=SPINPY_TEST_PRIVATE,
    )
)
print(resp)
resp = kc.temp(
    spin.KeyTempRequest(
        public=SPINPY_TEST_PUBLIC,
        private=SPINPY_TEST_PRIVATE,
        duration=60000000000,
    )
)
print(resp)

dc = sp.DirServerHTTPClient()
resp = dc.lookup(
    spin.DirLookupRequest(
        public=SPINPY_TEST_PUBLIC,
        private=SPINPY_TEST_PRIVATE,
        citizen=SPINPY_TEST_CITIZEN,
        path="/",
    )
)
print(resp)
resp = dc.tree(
    spin.DirTreeRequest(
        public=SPINPY_TEST_PUBLIC,
        private=SPINPY_TEST_PRIVATE,
        citizen=SPINPY_TEST_CITIZEN,
        path="/",
        level=1,
    )
)
print(resp)

resp = dc.apply(
    spin.DirApplyRequest(
        public=SPINPY_TEST_PUBLIC,
        private=SPINPY_TEST_PRIVATE,
        ops=[
            spin.DirOp(
                operation=spin.PutDirOperation,
                dir_entry=spin.DirEntry(
                    type=spin.EntryDir,
                    citizen="",
                    path="/test",
                    sequence=spin.SeqIgnore,
                ),
            )
        ],
    )
)
print(resp)

bc = sp.BitServerHTTPClient()
data = "Asdf".encode()
resp = bc.apply(
    spin.BitApplyRequest(
        public=SPINPY_TEST_PUBLIC,
        private=SPINPY_TEST_PRIVATE,
        ops=[
            spin.BitOp(
                operation=spin.PutBitOperation,
                ref=spin.SHA256(data),
                bytes=data,
            )
        ],
    )
)
print(resp)



import os

from spinsrv import spinfs

if "SPINPY_TEST_PUBLIC" not in os.environ:
    raise ValueError("must specify SPINPY_TEST_PUBLIC environment variable")
if "SPINPY_TEST_PRIVATE" not in os.environ:
    raise ValueError("must specify SPINPY_TEST_PRIVATE environment variable")
if "SPINPY_TEST_CITIZEN" not in os.environ:
    raise ValueError("must specify SPINPY_TEST_CITIZEN environment variable")
SPINPY_TEST_PUBLIC = os.environ["SPINPY_TEST_PUBLIC"]
SPINPY_TEST_PRIVATE = os.environ["SPINPY_TEST_PRIVATE"]
SPINPY_TEST_CITIZEN = os.environ["SPINPY_TEST_CITIZEN"]

print(
    "THIS IS SMOKE TEST 2: IT TESTS (primarily directory related) functionality of src/spin/spinfs.py"
)

fs = spinfs.SpinFileSystem(
    spinfs.Config(
        public=SPINPY_TEST_PUBLIC,
        private=SPINPY_TEST_PRIVATE,
        citizen=SPINPY_TEST_CITIZEN,
    )
)

resp = fs.ls("/")
print(resp)

resp = fs.makedirs("/this/a/test")
print(resp)

resp = fs.rm_file("/this/a/test")
print(resp)

resp = fs.touch("/this/a/test.txt")
print(resp)

assert fs.exists("/this/a/test.txt")
assert fs.isfile("/this/a/test.txt")
assert fs.cat("/this/a/test.txt") == b""
resp = fs.copy("/this/a/test.txt", "/this/a/copy.txt")
print(resp)
out = {x: True for x in fs.ls("/this/a", detail=False)}
assert "/this/a/test.txt" in out, f"out was {out}"
assert "/this/a/copy.txt" in out, f"out was {out}"
resp = fs.rm("/this/a", recursive=True)
print(resp)



"""
    pygments.console
    ~~~~~~~~~~~~~~~~

    Format colored console output.

    :copyright: Copyright 2006-2022 by the Pygments team, see AUTHORS.
    :license: BSD, see LICENSE for details.
"""

esc = "\x1b["

codes = {}
codes[""] = ""
codes["reset"] = esc + "39;49;00m"

codes["bold"] = esc + "01m"
codes["faint"] = esc + "02m"
codes["standout"] = esc + "03m"
codes["underline"] = esc + "04m"
codes["blink"] = esc + "05m"
codes["overline"] = esc + "06m"

dark_colors = ["black", "red", "green", "yellow", "blue", "magenta", "cyan", "gray"]
light_colors = [
    "brightblack",
    "brightred",
    "brightgreen",
    "brightyellow",
    "brightblue",
    "brightmagenta",
    "brightcyan",
    "white",
]

x = 30
for d, l in zip(dark_colors, light_colors):
    codes[d] = esc + "%im" % x
    codes[l] = esc + "%im" % (60 + x)
    x += 1

del d, l, x

codes["white"] = codes["bold"]


def reset_color():
    return codes["reset"]


def colorize(color_key, text):
    return codes[color_key] + text + codes["reset"]


def ansiformat(attr, text):
    """
    Format ``text`` with a color and/or some attributes::

        color       normal color
        *color*     bold color
        _color_     underlined color
        +color+     blinking color
    """
    result = []
    if attr[:1] == attr[-1:] == "+":
        result.append(codes["blink"])
        attr = attr[1:-1]
    if attr[:1] == attr[-1:] == "*":
        result.append(codes["bold"])
        attr = attr[1:-1]
    if attr[:1] == attr[-1:] == "_":
        result.append(codes["underline"])
        attr = attr[1:-1]
    result.append(codes[attr])
    result.append(text)
    result.append(codes["reset"])
    return "".join(result)



import inspect
import types
import typing as t
from functools import update_wrapper
from gettext import gettext as _

from .core import Argument
from .core import Command
from .core import Context
from .core import Group
from .core import Option
from .core import Parameter
from .globals import get_current_context
from .utils import echo

F = t.TypeVar("F", bound=t.Callable[..., t.Any])
FC = t.TypeVar("FC", bound=t.Union[t.Callable[..., t.Any], Command])


def pass_context(f: F) -> F:
    """Marks a callback as wanting to receive the current context
    object as first argument.
    """

    def new_func(*args, **kwargs):  # type: ignore
        return f(get_current_context(), *args, **kwargs)

    return update_wrapper(t.cast(F, new_func), f)


def pass_obj(f: F) -> F:
    """Similar to :func:`pass_context`, but only pass the object on the
    context onwards (:attr:`Context.obj`).  This is useful if that object
    represents the state of a nested system.
    """

    def new_func(*args, **kwargs):  # type: ignore
        return f(get_current_context().obj, *args, **kwargs)

    return update_wrapper(t.cast(F, new_func), f)


def make_pass_decorator(
    object_type: t.Type, ensure: bool = False
) -> "t.Callable[[F], F]":
    """Given an object type this creates a decorator that will work
    similar to :func:`pass_obj` but instead of passing the object of the
    current context, it will find the innermost context of type
    :func:`object_type`.

    This generates a decorator that works roughly like this::

        from functools import update_wrapper

        def decorator(f):
            @pass_context
            def new_func(ctx, *args, **kwargs):
                obj = ctx.find_object(object_type)
                return ctx.invoke(f, obj, *args, **kwargs)
            return update_wrapper(new_func, f)
        return decorator

    :param object_type: the type of the object to pass.
    :param ensure: if set to `True`, a new object will be created and
                   remembered on the context if it's not there yet.
    """

    def decorator(f: F) -> F:
        def new_func(*args, **kwargs):  # type: ignore
            ctx = get_current_context()

            if ensure:
                obj = ctx.ensure_object(object_type)
            else:
                obj = ctx.find_object(object_type)

            if obj is None:
                raise RuntimeError(
                    "Managed to invoke callback without a context"
                    f" object of type {object_type.__name__!r}"
                    " existing."
                )

            return ctx.invoke(f, obj, *args, **kwargs)

        return update_wrapper(t.cast(F, new_func), f)

    return decorator


def pass_meta_key(
    key: str, *, doc_description: t.Optional[str] = None
) -> "t.Callable[[F], F]":
    """Create a decorator that passes a key from
    :attr:`click.Context.meta` as the first argument to the decorated
    function.

    :param key: Key in ``Context.meta`` to pass.
    :param doc_description: Description of the object being passed,
        inserted into the decorator's docstring. Defaults to "the 'key'
        key from Context.meta".

    .. versionadded:: 8.0
    """

    def decorator(f: F) -> F:
        def new_func(*args, **kwargs):  # type: ignore
            ctx = get_current_context()
            obj = ctx.meta[key]
            return ctx.invoke(f, obj, *args, **kwargs)

        return update_wrapper(t.cast(F, new_func), f)

    if doc_description is None:
        doc_description = f"the {key!r} key from :attr:`click.Context.meta`"

    decorator.__doc__ = (
        f"Decorator that passes {doc_description} as the first argument"
        " to the decorated function."
    )
    return decorator


CmdType = t.TypeVar("CmdType", bound=Command)


@t.overload
def command(
    __func: t.Callable[..., t.Any],
) -> Command:
    ...


@t.overload
def command(
    name: t.Optional[str] = None,
    **attrs: t.Any,
) -> t.Callable[..., Command]:
    ...


@t.overload
def command(
    name: t.Optional[str] = None,
    cls: t.Type[CmdType] = ...,
    **attrs: t.Any,
) -> t.Callable[..., CmdType]:
    ...


def command(
    name: t.Union[str, t.Callable[..., t.Any], None] = None,
    cls: t.Optional[t.Type[Command]] = None,
    **attrs: t.Any,
) -> t.Union[Command, t.Callable[..., Command]]:
    r"""Creates a new :class:`Command` and uses the decorated function as
    callback.  This will also automatically attach all decorated
    :func:`option`\s and :func:`argument`\s as parameters to the command.

    The name of the command defaults to the name of the function with
    underscores replaced by dashes.  If you want to change that, you can
    pass the intended name as the first argument.

    All keyword arguments are forwarded to the underlying command class.
    For the ``params`` argument, any decorated params are appended to
    the end of the list.

    Once decorated the function turns into a :class:`Command` instance
    that can be invoked as a command line utility or be attached to a
    command :class:`Group`.

    :param name: the name of the command.  This defaults to the function
                 name with underscores replaced by dashes.
    :param cls: the command class to instantiate.  This defaults to
                :class:`Command`.

    .. versionchanged:: 8.1
        This decorator can be applied without parentheses.

    .. versionchanged:: 8.1
        The ``params`` argument can be used. Decorated params are
        appended to the end of the list.
    """

    func: t.Optional[t.Callable[..., t.Any]] = None

    if callable(name):
        func = name
        name = None
        assert cls is None, "Use 'command(cls=cls)(callable)' to specify a class."
        assert not attrs, "Use 'command(**kwargs)(callable)' to provide arguments."

    if cls is None:
        cls = Command

    def decorator(f: t.Callable[..., t.Any]) -> Command:
        if isinstance(f, Command):
            raise TypeError("Attempted to convert a callback into a command twice.")

        attr_params = attrs.pop("params", None)
        params = attr_params if attr_params is not None else []

        try:
            decorator_params = f.__click_params__  # type: ignore
        except AttributeError:
            pass
        else:
            del f.__click_params__  # type: ignore
            params.extend(reversed(decorator_params))

        if attrs.get("help") is None:
            attrs["help"] = f.__doc__

        cmd = cls(  # type: ignore[misc]
            name=name or f.__name__.lower().replace("_", "-"),  # type: ignore[arg-type]
            callback=f,
            params=params,
            **attrs,
        )
        cmd.__doc__ = f.__doc__
        return cmd

    if func is not None:
        return decorator(func)

    return decorator


@t.overload
def group(
    __func: t.Callable[..., t.Any],
) -> Group:
    ...


@t.overload
def group(
    name: t.Optional[str] = None,
    **attrs: t.Any,
) -> t.Callable[[F], Group]:
    ...


def group(
    name: t.Union[str, t.Callable[..., t.Any], None] = None, **attrs: t.Any
) -> t.Union[Group, t.Callable[[F], Group]]:
    """Creates a new :class:`Group` with a function as callback.  This
    works otherwise the same as :func:`command` just that the `cls`
    parameter is set to :class:`Group`.

    .. versionchanged:: 8.1
        This decorator can be applied without parentheses.
    """
    if attrs.get("cls") is None:
        attrs["cls"] = Group

    if callable(name):
        grp: t.Callable[[F], Group] = t.cast(Group, command(**attrs))
        return grp(name)

    return t.cast(Group, command(name, **attrs))


def _param_memo(f: FC, param: Parameter) -> None:
    if isinstance(f, Command):
        f.params.append(param)
    else:
        if not hasattr(f, "__click_params__"):
            f.__click_params__ = []  # type: ignore

        f.__click_params__.append(param)  # type: ignore


def argument(*param_decls: str, **attrs: t.Any) -> t.Callable[[FC], FC]:
    """Attaches an argument to the command.  All positional arguments are
    passed as parameter declarations to :class:`Argument`; all keyword
    arguments are forwarded unchanged (except ``cls``).
    This is equivalent to creating an :class:`Argument` instance manually
    and attaching it to the :attr:`Command.params` list.

    :param cls: the argument class to instantiate.  This defaults to
                :class:`Argument`.
    """

    def decorator(f: FC) -> FC:
        ArgumentClass = attrs.pop("cls", None) or Argument
        _param_memo(f, ArgumentClass(param_decls, **attrs))
        return f

    return decorator


def option(*param_decls: str, **attrs: t.Any) -> t.Callable[[FC], FC]:
    """Attaches an option to the command.  All positional arguments are
    passed as parameter declarations to :class:`Option`; all keyword
    arguments are forwarded unchanged (except ``cls``).
    This is equivalent to creating an :class:`Option` instance manually
    and attaching it to the :attr:`Command.params` list.

    :param cls: the option class to instantiate.  This defaults to
                :class:`Option`.
    """

    def decorator(f: FC) -> FC:
        # Issue 926, copy attrs, so pre-defined options can re-use the same cls=
        option_attrs = attrs.copy()
        OptionClass = option_attrs.pop("cls", None) or Option
        _param_memo(f, OptionClass(param_decls, **option_attrs))
        return f

    return decorator


def confirmation_option(*param_decls: str, **kwargs: t.Any) -> t.Callable[[FC], FC]:
    """Add a ``--yes`` option which shows a prompt before continuing if
    not passed. If the prompt is declined, the program will exit.

    :param param_decls: One or more option names. Defaults to the single
        value ``"--yes"``.
    :param kwargs: Extra arguments are passed to :func:`option`.
    """

    def callback(ctx: Context, param: Parameter, value: bool) -> None:
        if not value:
            ctx.abort()

    if not param_decls:
        param_decls = ("--yes",)

    kwargs.setdefault("is_flag", True)
    kwargs.setdefault("callback", callback)
    kwargs.setdefault("expose_value", False)
    kwargs.setdefault("prompt", "Do you want to continue?")
    kwargs.setdefault("help", "Confirm the action without prompting.")
    return option(*param_decls, **kwargs)


def password_option(*param_decls: str, **kwargs: t.Any) -> t.Callable[[FC], FC]:
    """Add a ``--password`` option which prompts for a password, hiding
    input and asking to enter the value again for confirmation.

    :param param_decls: One or more option names. Defaults to the single
        value ``"--password"``.
    :param kwargs: Extra arguments are passed to :func:`option`.
    """
    if not param_decls:
        param_decls = ("--password",)

    kwargs.setdefault("prompt", True)
    kwargs.setdefault("confirmation_prompt", True)
    kwargs.setdefault("hide_input", True)
    return option(*param_decls, **kwargs)


def version_option(
    version: t.Optional[str] = None,
    *param_decls: str,
    package_name: t.Optional[str] = None,
    prog_name: t.Optional[str] = None,
    message: t.Optional[str] = None,
    **kwargs: t.Any,
) -> t.Callable[[FC], FC]:
    """Add a ``--version`` option which immediately prints the version
    number and exits the program.

    If ``version`` is not provided, Click will try to detect it using
    :func:`importlib.metadata.version` to get the version for the
    ``package_name``. On Python < 3.8, the ``importlib_metadata``
    backport must be installed.

    If ``package_name`` is not provided, Click will try to detect it by
    inspecting the stack frames. This will be used to detect the
    version, so it must match the name of the installed package.

    :param version: The version number to show. If not provided, Click
        will try to detect it.
    :param param_decls: One or more option names. Defaults to the single
        value ``"--version"``.
    :param package_name: The package name to detect the version from. If
        not provided, Click will try to detect it.
    :param prog_name: The name of the CLI to show in the message. If not
        provided, it will be detected from the command.
    :param message: The message to show. The values ``%(prog)s``,
        ``%(package)s``, and ``%(version)s`` are available. Defaults to
        ``"%(prog)s, version %(version)s"``.
    :param kwargs: Extra arguments are passed to :func:`option`.
    :raise RuntimeError: ``version`` could not be detected.

    .. versionchanged:: 8.0
        Add the ``package_name`` parameter, and the ``%(package)s``
        value for messages.

    .. versionchanged:: 8.0
        Use :mod:`importlib.metadata` instead of ``pkg_resources``. The
        version is detected based on the package name, not the entry
        point name. The Python package name must match the installed
        package name, or be passed with ``package_name=``.
    """
    if message is None:
        message = _("%(prog)s, version %(version)s")

    if version is None and package_name is None:
        frame = inspect.currentframe()
        f_back = frame.f_back if frame is not None else None
        f_globals = f_back.f_globals if f_back is not None else None
        # break reference cycle
        # https://docs.python.org/3/library/inspect.html#the-interpreter-stack
        del frame

        if f_globals is not None:
            package_name = f_globals.get("__name__")

            if package_name == "__main__":
                package_name = f_globals.get("__package__")

            if package_name:
                package_name = package_name.partition(".")[0]

    def callback(ctx: Context, param: Parameter, value: bool) -> None:
        if not value or ctx.resilient_parsing:
            return

        nonlocal prog_name
        nonlocal version

        if prog_name is None:
            prog_name = ctx.find_root().info_name

        if version is None and package_name is not None:
            metadata: t.Optional[types.ModuleType]

            try:
                from importlib import metadata  # type: ignore
            except ImportError:
                # Python < 3.8
                import importlib_metadata as metadata  # type: ignore

            try:
                version = metadata.version(package_name)  # type: ignore
            except metadata.PackageNotFoundError:  # type: ignore
                raise RuntimeError(
                    f"{package_name!r} is not installed. Try passing"
                    " 'package_name' instead."
                ) from None

        if version is None:
            raise RuntimeError(
                f"Could not determine the version for {package_name!r} automatically."
            )

        echo(
            t.cast(str, message)
            % {"prog": prog_name, "package": package_name, "version": version},
            color=ctx.color,
        )
        ctx.exit()

    if not param_decls:
        param_decls = ("--version",)

    kwargs.setdefault("is_flag", True)
    kwargs.setdefault("expose_value", False)
    kwargs.setdefault("is_eager", True)
    kwargs.setdefault("help", _("Show the version and exit."))
    kwargs["callback"] = callback
    return option(*param_decls, **kwargs)


def help_option(*param_decls: str, **kwargs: t.Any) -> t.Callable[[FC], FC]:
    """Add a ``--help`` option which immediately prints the help page
    and exits the program.

    This is usually unnecessary, as the ``--help`` option is added to
    each command automatically unless ``add_help_option=False`` is
    passed.

    :param param_decls: One or more option names. Defaults to the single
        value ``"--help"``.
    :param kwargs: Extra arguments are passed to :func:`option`.
    """

    def callback(ctx: Context, param: Parameter, value: bool) -> None:
        if not value or ctx.resilient_parsing:
            return

        echo(ctx.get_help(), color=ctx.color)
        ctx.exit()

    if not param_decls:
        param_decls = ("--help",)

    kwargs.setdefault("is_flag", True)
    kwargs.setdefault("expose_value", False)
    kwargs.setdefault("is_eager", True)
    kwargs.setdefault("help", _("Show this message and exit."))
    kwargs["callback"] = callback
    return option(*param_decls, **kwargs)



"""
This module provides an object oriented interface for pattern matching
of files.
"""

from collections.abc import Collection as CollectionType
from itertools import zip_longest
from os import PathLike
from typing import (
    AnyStr,
    Callable,
    Collection,
    Iterable,
    Iterator,
    Optional,
    TYPE_CHECKING,
    Union,
)

from . import util
from .pattern import Pattern
from .util import TreeEntry, _filter_patterns, _is_iterable, match_file, normalize_file


class PathSpec(object):
    """
    The :class:`PathSpec` class is a wrapper around a list of compiled
    :class:`.Pattern` instances.
    """

    def __init__(self, patterns: Iterable[Pattern]) -> None:
        """
        Initializes the :class:`PathSpec` instance.

        *patterns* (:class:`~collections.abc.Collection` or :class:`~collections.abc.Iterable`)
        yields each compiled pattern (:class:`.Pattern`).
        """

        self.patterns = (
            patterns if isinstance(patterns, CollectionType) else list(patterns)
        )
        """
		*patterns* (:class:`~collections.abc.Collection` of :class:`.Pattern`)
		contains the compiled patterns.
		"""

    def __eq__(self, other: "Self") -> bool:
        """
        Tests the equality of this path-spec with *other* (:class:`PathSpec`)
        by comparing their :attr:`~PathSpec.patterns` attributes.
        """
        if isinstance(other, PathSpec):
            paired_patterns = zip_longest(self.patterns, other.patterns)
            return all(a == b for a, b in paired_patterns)
        else:
            return NotImplemented

    def __len__(self) -> int:
        """
        Returns the number of compiled patterns this path-spec contains
        (:class:`int`).
        """
        return len(self.patterns)

    def __add__(self, other: "Self") -> "Self":
        """
        Combines the :attr:`Pathspec.patterns` patterns from two
        :class:`PathSpec` instances.
        """
        if isinstance(other, PathSpec):
            return self.__class__(self.patterns + other.patterns)
        else:
            return NotImplemented

    def __iadd__(self, other: "Self") -> "Self":
        """
        Adds the :attr:`Pathspec.patterns` patterns from one :class:`PathSpec`
        instance to this instance.
        """
        if isinstance(other, PathSpec):
            self.patterns += other.patterns
            return self
        else:
            return NotImplemented

    @classmethod
    def from_lines(
        cls,
        pattern_factory: Union[str, Callable[[AnyStr], Pattern]],
        lines: Iterable[AnyStr],
    ) -> "Self":
        """
        Compiles the pattern lines.

        *pattern_factory* can be either the name of a registered pattern
        factory (:class:`str`), or a :class:`~collections.abc.Callable` used
        to compile patterns. It must accept an uncompiled pattern (:class:`str`)
        and return the compiled pattern (:class:`.Pattern`).

        *lines* (:class:`~collections.abc.Iterable`) yields each uncompiled
        pattern (:class:`str`). This simply has to yield each line so it can
        be a :class:`io.TextIOBase` (e.g., from :func:`open` or
        :class:`io.StringIO`) or the result from :meth:`str.splitlines`.

        Returns the :class:`PathSpec` instance.
        """
        if isinstance(pattern_factory, str):
            pattern_factory = util.lookup_pattern(pattern_factory)

        if not callable(pattern_factory):
            raise TypeError(f"pattern_factory:{pattern_factory!r} is not callable.")

        if not _is_iterable(lines):
            raise TypeError(f"lines:{lines!r} is not an iterable.")

        patterns = [pattern_factory(line) for line in lines if line]
        return cls(patterns)

    def match_entries(
        self,
        entries: Iterable[TreeEntry],
        separators: Optional[Collection[str]] = None,
    ) -> Iterator[TreeEntry]:
        """
        Matches the entries to this path-spec.

        *entries* (:class:`~collections.abc.Iterable` of :class:`~util.TreeEntry`)
        contains the entries to be matched against :attr:`self.patterns <PathSpec.patterns>`.

        *separators* (:class:`~collections.abc.Collection` of :class:`str`;
        or :data:`None`) optionally contains the path separators to
        normalize. See :func:`~pathspec.util.normalize_file` for more
        information.

        Returns the matched entries (:class:`~collections.abc.Iterator` of
        :class:`~util.TreeEntry`).
        """
        if not _is_iterable(entries):
            raise TypeError(f"entries:{entries!r} is not an iterable.")

        use_patterns = _filter_patterns(self.patterns)
        for entry in entries:
            norm_file = normalize_file(entry.path, separators)
            if self._match_file(use_patterns, norm_file):
                yield entry

    # Match files using the `match_file()` utility function. Subclasses
    # may override this method as an instance method. It does not have to
    # be a static method.
    _match_file = staticmethod(match_file)

    def match_file(
        self,
        file: Union[str, PathLike],
        separators: Optional[Collection[str]] = None,
    ) -> bool:
        """
        Matches the file to this path-spec.

        *file* (:class:`str` or :class:`os.PathLike`) is the file path to be
        matched against :attr:`self.patterns <PathSpec.patterns>`.

        *separators* (:class:`~collections.abc.Collection` of :class:`str`)
        optionally contains the path separators to normalize. See
        :func:`~pathspec.util.normalize_file` for more information.

        Returns :data:`True` if *file* matched; otherwise, :data:`False`.
        """
        norm_file = util.normalize_file(file, separators=separators)
        return self._match_file(self.patterns, norm_file)

    def match_files(
        self,
        files: Iterable[Union[str, PathLike]],
        separators: Optional[Collection[str]] = None,
    ) -> Iterator[Union[str, PathLike]]:
        """
        Matches the files to this path-spec.

        *files* (:class:`~collections.abc.Iterable` of :class:`str` or
        :class:`os.PathLike`) contains the file paths to be matched against
        :attr:`self.patterns <PathSpec.patterns>`.

        *separators* (:class:`~collections.abc.Collection` of :class:`str`;
        or :data:`None`) optionally contains the path separators to
        normalize. See :func:`~pathspec.util.normalize_file` for more
        information.

        Returns the matched files (:class:`~collections.abc.Iterator` of
        :class:`str` or :class:`os.PathLike`).
        """
        if not _is_iterable(files):
            raise TypeError(f"files:{files!r} is not an iterable.")

        use_patterns = _filter_patterns(self.patterns)
        for orig_file in files:
            norm_file = normalize_file(orig_file, separators)
            if self._match_file(use_patterns, norm_file):
                yield orig_file

    def match_tree_entries(
        self,
        root: Union[str, PathLike],
        on_error: Optional[Callable] = None,
        follow_links: Optional[bool] = None,
    ) -> Iterator[TreeEntry]:
        """
        Walks the specified root path for all files and matches them to this
        path-spec.

        *root* (:class:`str` or :class:`os.PathLike`) is the root directory
        to search.

        *on_error* (:class:`~collections.abc.Callable` or :data:`None`)
        optionally is the error handler for file-system exceptions. See
        :func:`~pathspec.util.iter_tree_entries` for more information.

        *follow_links* (:class:`bool` or :data:`None`) optionally is whether
        to walk symbolic links that resolve to directories. See
        :func:`~pathspec.util.iter_tree_files` for more information.

        Returns the matched files (:class:`~collections.abc.Iterator` of
        :class:`.TreeEntry`).
        """
        entries = util.iter_tree_entries(
            root, on_error=on_error, follow_links=follow_links
        )
        yield from self.match_entries(entries)

    def match_tree_files(
        self,
        root: Union[str, PathLike],
        on_error: Optional[Callable] = None,
        follow_links: Optional[bool] = None,
    ) -> Iterator[str]:
        """
        Walks the specified root path for all files and matches them to this
        path-spec.

        *root* (:class:`str` or :class:`os.PathLike`) is the root directory
        to search for files.

        *on_error* (:class:`~collections.abc.Callable` or :data:`None`)
        optionally is the error handler for file-system exceptions. See
        :func:`~pathspec.util.iter_tree_files` for more information.

        *follow_links* (:class:`bool` or :data:`None`) optionally is whether
        to walk symbolic links that resolve to directories. See
        :func:`~pathspec.util.iter_tree_files` for more information.

        Returns the matched files (:class:`~collections.abc.Iterable` of
        :class:`str`).
        """
        files = util.iter_tree_files(root, on_error=on_error, follow_links=follow_links)
        yield from self.match_files(files)

    # Alias `match_tree_files()` as `match_tree()` for backward
    # compatibility before v0.3.2.
    match_tree = match_tree_files


if TYPE_CHECKING:
    try:
        from typing import Self
    except ImportError:
        try:
            from typing_extensions import Self
        except ImportError:
            Self = PathSpec



"""
    pygments.formatters.pangomarkup
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    Formatter for Pango markup output.

    :copyright: Copyright 2006-2022 by the Pygments team, see AUTHORS.
    :license: BSD, see LICENSE for details.
"""

from pip._vendor.pygments.formatter import Formatter


__all__ = ["PangoMarkupFormatter"]


_escape_table = {
    ord("&"): "&amp;",
    ord("<"): "&lt;",
}


def escape_special_chars(text, table=_escape_table):
    """Escape & and < for Pango Markup."""
    return text.translate(table)


class PangoMarkupFormatter(Formatter):
    """
    Format tokens as Pango Markup code. It can then be rendered to an SVG.

    .. versionadded:: 2.9
    """

    name = "Pango Markup"
    aliases = ["pango", "pangomarkup"]
    filenames = []

    def __init__(self, **options):
        Formatter.__init__(self, **options)

        self.styles = {}

        for token, style in self.style:
            start = ""
            end = ""
            if style["color"]:
                start += '<span fgcolor="#%s">' % style["color"]
                end = "</span>" + end
            if style["bold"]:
                start += "<b>"
                end = "</b>" + end
            if style["italic"]:
                start += "<i>"
                end = "</i>" + end
            if style["underline"]:
                start += "<u>"
                end = "</u>" + end
            self.styles[token] = (start, end)

    def format_unencoded(self, tokensource, outfile):
        lastval = ""
        lasttype = None

        outfile.write("<tt>")

        for ttype, value in tokensource:
            while ttype not in self.styles:
                ttype = ttype.parent
            if ttype == lasttype:
                lastval += escape_special_chars(value)
            else:
                if lastval:
                    stylebegin, styleend = self.styles[lasttype]
                    outfile.write(stylebegin + lastval + styleend)
                lastval = escape_special_chars(value)
                lasttype = ttype

        if lastval:
            stylebegin, styleend = self.styles[lasttype]
            outfile.write(stylebegin + lastval + styleend)

        outfile.write("</tt>")



"""
This module provides :class:`.GitIgnoreSpec` which replicates
*.gitignore* behavior.
"""

from typing import AnyStr, Callable, Collection, Iterable, TYPE_CHECKING, Union

from .pathspec import PathSpec
from .pattern import Pattern
from .patterns.gitwildmatch import (
    GitWildMatchPattern,
    GitWildMatchPatternError,
    _DIR_MARK,
)
from .util import _is_iterable


class GitIgnoreSpec(PathSpec):
    """
    The :class:`GitIgnoreSpec` class extends :class:`PathSpec` to
    replicate *.gitignore* behavior.
    """

    def __eq__(self, other: "Self") -> bool:
        """
        Tests the equality of this gitignore-spec with *other*
        (:class:`GitIgnoreSpec`) by comparing their :attr:`~PathSpec.patterns`
        attributes. A non-:class:`GitIgnoreSpec` will not compare equal.
        """
        if isinstance(other, GitIgnoreSpec):
            return super().__eq__(other)
        elif isinstance(other, PathSpec):
            return False
        else:
            return NotImplemented

    @classmethod
    def from_lines(
        cls,
        lines: Iterable[AnyStr],
        pattern_factory: Union[str, Callable[[AnyStr], Pattern], None] = None,
    ) -> "Self":
        """
        Compiles the pattern lines.

        *lines* (:class:`~collections.abc.Iterable`) yields each uncompiled
        pattern (:class:`str`). This simply has to yield each line so it can
        be a :class:`io.TextIOBase` (e.g., from :func:`open` or
        :class:`io.StringIO`) or the result from :meth:`str.splitlines`.

        *pattern_factory* can be :data:`None`, the name of a registered
        pattern factory (:class:`str`), or a :class:`~collections.abc.Callable`
        used to compile patterns. The callable must accept an uncompiled
        pattern (:class:`str`) and return the compiled pattern (:class:`.Pattern`).
        Default is :data:`None` for :class:`.GitWildMatchPattern`).

        Returns the :class:`GitIgnoreSpec` instance.
        """
        if pattern_factory is None:
            pattern_factory = GitWildMatchPattern

        elif (isinstance(lines, str) or callable(lines)) and _is_iterable(
            pattern_factory
        ):
            # Support reversed order of arguments from PathSpec.
            pattern_factory, lines = lines, pattern_factory

        self = super().from_lines(pattern_factory, lines)
        return self  # type: ignore

    @staticmethod
    def _match_file(
        patterns: Collection[GitWildMatchPattern],
        file: str,
    ) -> bool:
        """
        Matches the file to the patterns.

        .. NOTE:: Subclasses of :class:`.PathSpec` may override this
           method as an instance method. It does not have to be a static
           method.

        *patterns* (:class:`~collections.abc.Iterable` of :class:`~pathspec.pattern.Pattern`)
        contains the patterns to use.

        *file* (:class:`str`) is the normalized file path to be matched
        against *patterns*.

        Returns :data:`True` if *file* matched; otherwise, :data:`False`.
        """
        out_matched = False
        out_priority = 0
        for pattern in patterns:
            if pattern.include is not None:
                match = pattern.match_file(file)
                if match is not None:
                    # Pattern matched.

                    # Check for directory marker.
                    try:
                        dir_mark = match.match.group(_DIR_MARK)
                    except IndexError as e:
                        # NOTICE: The exact content of this error message is subject
                        # to change.
                        raise GitWildMatchPatternError(
                            (
                                f"Invalid git pattern: directory marker regex group is missing. "
                                f"Debug: file={file!r} regex={pattern.regex!r} "
                                f"group={_DIR_MARK!r} match={match.match!r}."
                            )
                        ) from e

                    if dir_mark:
                        # Pattern matched by a directory pattern.
                        priority = 1
                    else:
                        # Pattern matched by a file pattern.
                        priority = 2

                    if priority >= out_priority:
                        out_matched = pattern.include
                        out_priority = priority

        return out_matched


if TYPE_CHECKING:
    try:
        from typing import Self
    except ImportError:
        try:
            from typing_extensions import Self
        except ImportError:
            Self = GitIgnoreSpec



"""
This module provides utility methods for dealing with path-specs.
"""

import os
import os.path
import posixpath
import stat
import warnings
from collections.abc import Collection as CollectionType, Iterable as IterableType
from os import PathLike
from typing import (
    Any,
    AnyStr,
    Callable,
    Collection,
    Dict,
    Iterable,
    Iterator,
    List,
    Optional,
    Sequence,
    Set,
    Union,
)

from .pattern import Pattern

NORMALIZE_PATH_SEPS = [
    __sep for __sep in [os.sep, os.altsep] if __sep and __sep != posixpath.sep
]
"""
*NORMALIZE_PATH_SEPS* (:class:`list` of :class:`str`) contains the path
separators that need to be normalized to the POSIX separator for the
current operating system. The separators are determined by examining
:data:`os.sep` and :data:`os.altsep`.
"""

_registered_patterns = {}
"""
*_registered_patterns* (:class:`dict`) maps a name (:class:`str`) to the
registered pattern factory (:class:`~collections.abc.Callable`).
"""


def detailed_match_files(
    patterns: Iterable[Pattern],
    files: Iterable[str],
    all_matches: Optional[bool] = None,
) -> Dict[str, "MatchDetail"]:
    """
    Matches the files to the patterns, and returns which patterns matched
    the files.

    *patterns* (:class:`~collections.abc.Iterable` of :class:`~pathspec.pattern.Pattern`)
    contains the patterns to use.

    *files* (:class:`~collections.abc.Iterable` of :class:`str`) contains
    the normalized file paths to be matched against *patterns*.

    *all_matches* (:class:`boot` or :data:`None`) is whether to return all
    matches patterns (:data:`True`), or only the last matched pattern
    (:data:`False`). Default is :data:`None` for :data:`False`.

    Returns the matched files (:class:`dict`) which maps each matched file
    (:class:`str`) to the patterns that matched in order (:class:`.MatchDetail`).
    """
    all_files = files if isinstance(files, CollectionType) else list(files)
    return_files = {}
    for pattern in patterns:
        if pattern.include is not None:
            result_files = pattern.match(
                all_files
            )  # TODO: Replace with `.match_file()`.
            if pattern.include:
                # Add files and record pattern.
                for result_file in result_files:
                    if result_file in return_files:
                        if all_matches:
                            return_files[result_file].patterns.append(pattern)
                        else:
                            return_files[result_file].patterns[0] = pattern
                    else:
                        return_files[result_file] = MatchDetail([pattern])

            else:
                # Remove files.
                for file in result_files:
                    del return_files[file]

    return return_files


def _filter_patterns(patterns: Iterable[Pattern]) -> List[Pattern]:
    """
    Filters out null-patterns.

    *patterns* (:class:`Iterable` of :class:`.Pattern`) contains the
    patterns.

    Returns the patterns (:class:`list` of :class:`.Pattern`).
    """
    return [__pat for __pat in patterns if __pat.include is not None]


def _is_iterable(value: Any) -> bool:
    """
    Check whether the value is an iterable (excludes strings).

    *value* is the value to check,

    Returns whether *value* is a iterable (:class:`bool`).
    """
    return isinstance(value, IterableType) and not isinstance(value, (str, bytes))


def iter_tree_entries(
    root: Union[str, PathLike],
    on_error: Optional[Callable] = None,
    follow_links: Optional[bool] = None,
) -> Iterator["TreeEntry"]:
    """
    Walks the specified directory for all files and directories.

    *root* (:class:`str` or :class:`os.PathLike`) is the root directory to
    search.

    *on_error* (:class:`~collections.abc.Callable` or :data:`None`)
    optionally is the error handler for file-system exceptions. It will be
    called with the exception (:exc:`OSError`). Reraise the exception to
    abort the walk. Default is :data:`None` to ignore file-system
    exceptions.

    *follow_links* (:class:`bool` or :data:`None`) optionally is whether
    to walk symbolic links that resolve to directories. Default is
    :data:`None` for :data:`True`.

    Raises :exc:`RecursionError` if recursion is detected.

    Returns an :class:`~collections.abc.Iterator` yielding each file or
    directory entry (:class:`.TreeEntry`) relative to *root*.
    """
    if on_error is not None and not callable(on_error):
        raise TypeError(f"on_error:{on_error!r} is not callable.")

    if follow_links is None:
        follow_links = True

    yield from _iter_tree_entries_next(
        os.path.abspath(root), "", {}, on_error, follow_links
    )


def _iter_tree_entries_next(
    root_full: str,
    dir_rel: str,
    memo: Dict[str, str],
    on_error: Callable,
    follow_links: bool,
) -> Iterator["TreeEntry"]:
    """
    Scan the directory for all descendant files.

    *root_full* (:class:`str`) the absolute path to the root directory.

    *dir_rel* (:class:`str`) the path to the directory to scan relative to
    *root_full*.

    *memo* (:class:`dict`) keeps track of ancestor directories
    encountered. Maps each ancestor real path (:class:`str`) to relative
    path (:class:`str`).

    *on_error* (:class:`~collections.abc.Callable` or :data:`None`)
    optionally is the error handler for file-system exceptions.

    *follow_links* (:class:`bool`) is whether to walk symbolic links that
    resolve to directories.

    Yields each entry (:class:`.TreeEntry`).
    """
    dir_full = os.path.join(root_full, dir_rel)
    dir_real = os.path.realpath(dir_full)

    # Remember each encountered ancestor directory and its canonical
    # (real) path. If a canonical path is encountered more than once,
    # recursion has occurred.
    if dir_real not in memo:
        memo[dir_real] = dir_rel
    else:
        raise RecursionError(
            real_path=dir_real, first_path=memo[dir_real], second_path=dir_rel
        )

    with os.scandir(dir_full) as scan_iter:
        node_ent: os.DirEntry
        for node_ent in scan_iter:
            node_rel = os.path.join(dir_rel, node_ent.name)

            # Inspect child node.
            try:
                node_lstat = node_ent.stat(follow_symlinks=False)
            except OSError as e:
                if on_error is not None:
                    on_error(e)
                continue

            if node_ent.is_symlink():
                # Child node is a link, inspect the target node.
                try:
                    node_stat = node_ent.stat()
                except OSError as e:
                    if on_error is not None:
                        on_error(e)
                    continue
            else:
                node_stat = node_lstat

            if node_ent.is_dir(follow_symlinks=follow_links):
                # Child node is a directory, recurse into it and yield its
                # descendant files.
                yield TreeEntry(node_ent.name, node_rel, node_lstat, node_stat)

                yield from _iter_tree_entries_next(
                    root_full, node_rel, memo, on_error, follow_links
                )

            elif node_ent.is_file() or node_ent.is_symlink():
                # Child node is either a file or an unfollowed link, yield it.
                yield TreeEntry(node_ent.name, node_rel, node_lstat, node_stat)

    # NOTE: Make sure to remove the canonical (real) path of the directory
    # from the ancestors memo once we are done with it. This allows the
    # same directory to appear multiple times. If this is not done, the
    # second occurrence of the directory will be incorrectly interpreted
    # as a recursion. See <https://github.com/cpburnz/python-path-specification/pull/7>.
    del memo[dir_real]


def iter_tree_files(
    root: Union[str, PathLike],
    on_error: Optional[Callable] = None,
    follow_links: Optional[bool] = None,
) -> Iterator[str]:
    """
    Walks the specified directory for all files.

    *root* (:class:`str` or :class:`os.PathLike`) is the root directory to
    search for files.

    *on_error* (:class:`~collections.abc.Callable` or :data:`None`)
    optionally is the error handler for file-system exceptions. It will be
    called with the exception (:exc:`OSError`). Reraise the exception to
    abort the walk. Default is :data:`None` to ignore file-system
    exceptions.

    *follow_links* (:class:`bool` or :data:`None`) optionally is whether
    to walk symbolic links that resolve to directories. Default is
    :data:`None` for :data:`True`.

    Raises :exc:`RecursionError` if recursion is detected.

    Returns an :class:`~collections.abc.Iterator` yielding the path to
    each file (:class:`str`) relative to *root*.
    """
    for entry in iter_tree_entries(root, on_error=on_error, follow_links=follow_links):
        if not entry.is_dir(follow_links):
            yield entry.path


def iter_tree(root, on_error=None, follow_links=None):
    """
    DEPRECATED: The :func:`.iter_tree` function is an alias for the
    :func:`.iter_tree_files` function.
    """
    warnings.warn(
        ("util.iter_tree() is deprecated. Use util.iter_tree_files() instead."),
        DeprecationWarning,
        stacklevel=2,
    )
    return iter_tree_files(root, on_error=on_error, follow_links=follow_links)


def lookup_pattern(name: str) -> Callable[[AnyStr], Pattern]:
    """
    Lookups a registered pattern factory by name.

    *name* (:class:`str`) is the name of the pattern factory.

    Returns the registered pattern factory (:class:`~collections.abc.Callable`).
    If no pattern factory is registered, raises :exc:`KeyError`.
    """
    return _registered_patterns[name]


def match_file(patterns: Iterable[Pattern], file: str) -> bool:
    """
    Matches the file to the patterns.

    *patterns* (:class:`~collections.abc.Iterable` of :class:`~pathspec.pattern.Pattern`)
    contains the patterns to use.

    *file* (:class:`str`) is the normalized file path to be matched
    against *patterns*.

    Returns :data:`True` if *file* matched; otherwise, :data:`False`.
    """
    matched = False
    for pattern in patterns:
        if pattern.include is not None:
            if pattern.match_file(file) is not None:
                matched = pattern.include

    return matched


def match_files(
    patterns: Iterable[Pattern],
    files: Iterable[str],
) -> Set[str]:
    """
    DEPRECATED: This is an old function no longer used. Use the :func:`.match_file`
    function with a loop for better results.

    Matches the files to the patterns.

    *patterns* (:class:`~collections.abc.Iterable` of :class:`~pathspec.pattern.Pattern`)
    contains the patterns to use.

    *files* (:class:`~collections.abc.Iterable` of :class:`str`) contains
    the normalized file paths to be matched against *patterns*.

    Returns the matched files (:class:`set` of :class:`str`).
    """
    warnings.warn(
        (
            "util.match_files() is deprecated. Use util.match_file() with a "
            "loop for better results."
        ),
        DeprecationWarning,
        stacklevel=2,
    )

    use_patterns = _filter_patterns(patterns)

    return_files = set()
    for file in files:
        if match_file(use_patterns, file):
            return_files.add(file)

    return return_files


def normalize_file(
    file: Union[str, PathLike],
    separators: Optional[Collection[str]] = None,
) -> str:
    """
    Normalizes the file path to use the POSIX path separator (i.e.,
    :data:`'/'`), and make the paths relative (remove leading :data:`'/'`).

    *file* (:class:`str` or :class:`os.PathLike`) is the file path.

    *separators* (:class:`~collections.abc.Collection` of :class:`str`; or
    :data:`None`) optionally contains the path separators to normalize.
    This does not need to include the POSIX path separator (:data:`'/'`),
    but including it will not affect the results. Default is :data:`None`
    for :data:`NORMALIZE_PATH_SEPS`. To prevent normalization, pass an
    empty container (e.g., an empty tuple :data:`()`).

    Returns the normalized file path (:class:`str`).
    """
    # Normalize path separators.
    if separators is None:
        separators = NORMALIZE_PATH_SEPS

    # Convert path object to string.
    norm_file = str(file)

    for sep in separators:
        norm_file = norm_file.replace(sep, posixpath.sep)

    if norm_file.startswith("/"):
        # Make path relative.
        norm_file = norm_file[1:]

    elif norm_file.startswith("./"):
        # Remove current directory prefix.
        norm_file = norm_file[2:]

    return norm_file


def normalize_files(
    files: Iterable[Union[str, PathLike]],
    separators: Optional[Collection[str]] = None,
) -> Dict[str, List[Union[str, PathLike]]]:
    """
    DEPRECATED: This function is no longer used. Use the :func:`.normalize_file`
    function with a loop for better results.

    Normalizes the file paths to use the POSIX path separator.

    *files* (:class:`~collections.abc.Iterable` of :class:`str` or
    :class:`os.PathLike`) contains the file paths to be normalized.

    *separators* (:class:`~collections.abc.Collection` of :class:`str`; or
    :data:`None`) optionally contains the path separators to normalize.
    See :func:`normalize_file` for more information.

    Returns a :class:`dict` mapping each normalized file path (:class:`str`)
    to the original file paths (:class:`list` of :class:`str` or
    :class:`os.PathLike`).
    """
    warnings.warn(
        (
            "util.normalize_files() is deprecated. Use util.normalize_file() "
            "with a loop for better results."
        ),
        DeprecationWarning,
        stacklevel=2,
    )

    norm_files = {}
    for path in files:
        norm_file = normalize_file(path, separators=separators)
        if norm_file in norm_files:
            norm_files[norm_file].append(path)
        else:
            norm_files[norm_file] = [path]

    return norm_files


def register_pattern(
    name: str,
    pattern_factory: Callable[[AnyStr], Pattern],
    override: Optional[bool] = None,
) -> None:
    """
    Registers the specified pattern factory.

    *name* (:class:`str`) is the name to register the pattern factory
    under.

    *pattern_factory* (:class:`~collections.abc.Callable`) is used to
    compile patterns. It must accept an uncompiled pattern (:class:`str`)
    and return the compiled pattern (:class:`.Pattern`).

    *override* (:class:`bool` or :data:`None`) optionally is whether to
    allow overriding an already registered pattern under the same name
    (:data:`True`), instead of raising an :exc:`AlreadyRegisteredError`
    (:data:`False`). Default is :data:`None` for :data:`False`.
    """
    if not isinstance(name, str):
        raise TypeError(f"name:{name!r} is not a string.")

    if not callable(pattern_factory):
        raise TypeError(f"pattern_factory:{pattern_factory!r} is not callable.")

    if name in _registered_patterns and not override:
        raise AlreadyRegisteredError(name, _registered_patterns[name])

    _registered_patterns[name] = pattern_factory


class AlreadyRegisteredError(Exception):
    """
    The :exc:`AlreadyRegisteredError` exception is raised when a pattern
    factory is registered under a name already in use.
    """

    def __init__(
        self,
        name: str,
        pattern_factory: Callable[[AnyStr], Pattern],
    ) -> None:
        """
        Initializes the :exc:`AlreadyRegisteredError` instance.

        *name* (:class:`str`) is the name of the registered pattern.

        *pattern_factory* (:class:`~collections.abc.Callable`) is the
        registered pattern factory.
        """
        super(AlreadyRegisteredError, self).__init__(name, pattern_factory)

    @property
    def message(self) -> str:
        """
        *message* (:class:`str`) is the error message.
        """
        return "{name!r} is already registered for pattern factory:{pattern_factory!r}.".format(
            name=self.name,
            pattern_factory=self.pattern_factory,
        )

    @property
    def name(self) -> str:
        """
        *name* (:class:`str`) is the name of the registered pattern.
        """
        return self.args[0]

    @property
    def pattern_factory(self) -> Callable[[AnyStr], Pattern]:
        """
        *pattern_factory* (:class:`~collections.abc.Callable`) is the
        registered pattern factory.
        """
        return self.args[1]


class RecursionError(Exception):
    """
    The :exc:`RecursionError` exception is raised when recursion is
    detected.
    """

    def __init__(
        self,
        real_path: str,
        first_path: str,
        second_path: str,
    ) -> None:
        """
        Initializes the :exc:`RecursionError` instance.

        *real_path* (:class:`str`) is the real path that recursion was
        encountered on.

        *first_path* (:class:`str`) is the first path encountered for
        *real_path*.

        *second_path* (:class:`str`) is the second path encountered for
        *real_path*.
        """
        super(RecursionError, self).__init__(real_path, first_path, second_path)

    @property
    def first_path(self) -> str:
        """
        *first_path* (:class:`str`) is the first path encountered for
        :attr:`self.real_path <RecursionError.real_path>`.
        """
        return self.args[1]

    @property
    def message(self) -> str:
        """
        *message* (:class:`str`) is the error message.
        """
        return "Real path {real!r} was encountered at {first!r} and then {second!r}.".format(
            real=self.real_path,
            first=self.first_path,
            second=self.second_path,
        )

    @property
    def real_path(self) -> str:
        """
        *real_path* (:class:`str`) is the real path that recursion was
        encountered on.
        """
        return self.args[0]

    @property
    def second_path(self) -> str:
        """
        *second_path* (:class:`str`) is the second path encountered for
        :attr:`self.real_path <RecursionError.real_path>`.
        """
        return self.args[2]


class MatchDetail(object):
    """
    The :class:`.MatchDetail` class contains information about
    """

    # Make the class dict-less.
    __slots__ = ("patterns",)

    def __init__(self, patterns: Sequence[Pattern]) -> None:
        """
        Initialize the :class:`.MatchDetail` instance.

        *patterns* (:class:`~collections.abc.Sequence` of :class:`~pathspec.pattern.Pattern`)
        contains the patterns that matched the file in the order they were
        encountered.
        """

        self.patterns = patterns
        """
		*patterns* (:class:`~collections.abc.Sequence` of :class:`~pathspec.pattern.Pattern`)
		contains the patterns that matched the file in the order they were
		encountered.
		"""


class TreeEntry(object):
    """
    The :class:`.TreeEntry` class contains information about a file-system
    entry.
    """

    # Make the class dict-less.
    __slots__ = ("_lstat", "name", "path", "_stat")

    def __init__(
        self,
        name: str,
        path: str,
        lstat: os.stat_result,
        stat: os.stat_result,
    ) -> None:
        """
        Initialize the :class:`.TreeEntry` instance.

        *name* (:class:`str`) is the base name of the entry.

        *path* (:class:`str`) is the relative path of the entry.

        *lstat* (:class:`os.stat_result`) is the stat result of the direct
        entry.

        *stat* (:class:`os.stat_result`) is the stat result of the entry,
        potentially linked.
        """

        self._lstat: os.stat_result = lstat
        """
		*_lstat* (:class:`os.stat_result`) is the stat result of the direct
		entry.
		"""

        self.name: str = name
        """
		*name* (:class:`str`) is the base name of the entry.
		"""

        self.path: str = path
        """
		*path* (:class:`str`) is the path of the entry.
		"""

        self._stat: os.stat_result = stat
        """
		*_stat* (:class:`os.stat_result`) is the stat result of the linked
		entry.
		"""

    def is_dir(self, follow_links: Optional[bool] = None) -> bool:
        """
        Get whether the entry is a directory.

        *follow_links* (:class:`bool` or :data:`None`) is whether to follow
        symbolic links. If this is :data:`True`, a symlink to a directory
        will result in :data:`True`. Default is :data:`None` for :data:`True`.

        Returns whether the entry is a directory (:class:`bool`).
        """
        if follow_links is None:
            follow_links = True

        node_stat = self._stat if follow_links else self._lstat
        return stat.S_ISDIR(node_stat.st_mode)

    def is_file(self, follow_links: Optional[bool] = None) -> bool:
        """
        Get whether the entry is a regular file.

        *follow_links* (:class:`bool` or :data:`None`) is whether to follow
        symbolic links. If this is :data:`True`, a symlink to a regular file
        will result in :data:`True`. Default is :data:`None` for :data:`True`.

        Returns whether the entry is a regular file (:class:`bool`).
        """
        if follow_links is None:
            follow_links = True

        node_stat = self._stat if follow_links else self._lstat
        return stat.S_ISREG(node_stat.st_mode)

    def is_symlink(self) -> bool:
        """
        Returns whether the entry is a symbolic link (:class:`bool`).
        """
        return stat.S_ISLNK(self._lstat.st_mode)

    def stat(self, follow_links: Optional[bool] = None) -> os.stat_result:
        """
        Get the cached stat result for the entry.

        *follow_links* (:class:`bool` or :data:`None`) is whether to follow
        symbolic links. If this is :data:`True`, the stat result of the
        linked file will be returned. Default is :data:`None` for :data:`True`.

        Returns that stat result (:class:`os.stat_result`).
        """
        if follow_links is None:
            follow_links = True

        return self._stat if follow_links else self._lstat



import errno
import io
import os
import secrets
import shutil
from contextlib import suppress
from functools import wraps

from fsspec.spec import AbstractFileSystem
from fsspec.utils import infer_storage_options, mirror_from


def wrap_exceptions(func):
    @wraps(func)
    def wrapper(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except OSError as exception:
            if not exception.args:
                raise

            message, *args = exception.args
            if isinstance(message, str) and "does not exist" in message:
                raise FileNotFoundError(errno.ENOENT, message) from exception
            else:
                raise

    return wrapper


PYARROW_VERSION = None


class ArrowFSWrapper(AbstractFileSystem):
    """FSSpec-compatible wrapper of pyarrow.fs.FileSystem.

    Parameters
    ----------
    fs : pyarrow.fs.FileSystem

    """

    root_marker = "/"

    def __init__(self, fs, **kwargs):
        from pyarrow import __version__

        global PYARROW_VERSION
        PYARROW_VERSION = tuple(map(int, __version__.split(".")))
        self.fs = fs
        super().__init__(**kwargs)

    @classmethod
    def _strip_protocol(cls, path):
        ops = infer_storage_options(path)
        path = ops["path"]
        if path.startswith("//"):
            # special case for "hdfs://path" (without the triple slash)
            path = path[1:]
        return path

    def ls(self, path, detail=False, **kwargs):
        path = self._strip_protocol(path)
        from pyarrow.fs import FileSelector

        entries = [
            self._make_entry(entry)
            for entry in self.fs.get_file_info(FileSelector(path))
        ]
        if detail:
            return entries
        else:
            return [entry["name"] for entry in entries]

    def info(self, path, **kwargs):
        path = self._strip_protocol(path)
        [info] = self.fs.get_file_info([path])
        return self._make_entry(info)

    def exists(self, path):
        path = self._strip_protocol(path)
        try:
            self.info(path)
        except FileNotFoundError:
            return False
        else:
            return True

    def _make_entry(self, info):
        from pyarrow.fs import FileType

        if info.type is FileType.Directory:
            kind = "directory"
        elif info.type is FileType.File:
            kind = "file"
        elif info.type is FileType.NotFound:
            raise FileNotFoundError(errno.ENOENT, os.strerror(errno.ENOENT), info.path)
        else:
            kind = "other"

        return {
            "name": info.path,
            "size": info.size,
            "type": kind,
            "mtime": info.mtime,
        }

    @wrap_exceptions
    def cp_file(self, path1, path2, **kwargs):
        path1 = self._strip_protocol(path1).rstrip("/")
        path2 = self._strip_protocol(path2).rstrip("/")

        with self._open(path1, "rb") as lstream:
            tmp_fname = f"{path2}.tmp.{secrets.token_hex(6)}"
            try:
                with self.open(tmp_fname, "wb") as rstream:
                    shutil.copyfileobj(lstream, rstream)
                self.fs.move(tmp_fname, path2)
            except BaseException:  # noqa
                with suppress(FileNotFoundError):
                    self.fs.delete_file(tmp_fname)
                raise

    @wrap_exceptions
    def mv(self, path1, path2, **kwargs):
        path1 = self._strip_protocol(path1).rstrip("/")
        path2 = self._strip_protocol(path2).rstrip("/")
        self.fs.move(path1, path2)

    mv_file = mv

    @wrap_exceptions
    def rm_file(self, path):
        path = self._strip_protocol(path)
        self.fs.delete_file(path)

    @wrap_exceptions
    def rm(self, path, recursive=False, maxdepth=None):
        path = self._strip_protocol(path).rstrip("/")
        if self.isdir(path):
            if recursive:
                self.fs.delete_dir(path)
            else:
                raise ValueError("Can't delete directories without recursive=False")
        else:
            self.fs.delete_file(path)

    @wrap_exceptions
    def _open(self, path, mode="rb", block_size=None, seekable=False, **kwargs):
        if mode == "rb":
            if seekable:
                method = self.fs.open_input_file
            else:
                method = self.fs.open_input_stream
        elif mode == "wb":
            method = self.fs.open_output_stream
        elif mode == "ab":
            method = self.fs.open_append_stream
        else:
            raise ValueError(f"unsupported mode for Arrow filesystem: {mode!r}")

        _kwargs = {}
        if mode != "rb" or not seekable:
            if PYARROW_VERSION[0] >= 4:
                # disable compression auto-detection
                _kwargs["compression"] = None
        stream = method(path, **_kwargs)

        return ArrowFile(self, stream, path, mode, block_size, **kwargs)

    @wrap_exceptions
    def mkdir(self, path, create_parents=True, **kwargs):
        path = self._strip_protocol(path)
        if create_parents:
            self.makedirs(path, exist_ok=True)
        else:
            self.fs.create_dir(path, recursive=False)

    @wrap_exceptions
    def makedirs(self, path, exist_ok=False):
        path = self._strip_protocol(path)
        self.fs.create_dir(path, recursive=True)

    @wrap_exceptions
    def rmdir(self, path):
        path = self._strip_protocol(path)
        self.fs.delete_dir(path)

    @wrap_exceptions
    def modified(self, path):
        path = self._strip_protocol(path)
        return self.fs.get_file_info(path).mtime


@mirror_from(
    "stream", ["read", "seek", "tell", "write", "readable", "writable", "close", "size"]
)
class ArrowFile(io.IOBase):
    def __init__(self, fs, stream, path, mode, block_size=None, **kwargs):
        self.path = path
        self.mode = mode

        self.fs = fs
        self.stream = stream

        self.blocksize = self.block_size = block_size
        self.kwargs = kwargs

    def __enter__(self):
        return self

    def __exit__(self, *args):
        return self.close()


class HadoopFileSystem(ArrowFSWrapper):
    """A wrapper on top of the pyarrow.fs.HadoopFileSystem
    to connect it's interface with fsspec"""

    protocol = "hdfs"

    def __init__(
        self,
        host="default",
        port=0,
        user=None,
        kerb_ticket=None,
        extra_conf=None,
        **kwargs,
    ):
        """

        Parameters
        ----------
        host: str
            Hostname, IP or "default" to try to read from Hadoop config
        port: int
            Port to connect on, or default from Hadoop config if 0
        user: str or None
            If given, connect as this username
        kerb_ticket: str or None
            If given, use this ticket for authentication
        extra_conf: None or dict
            Passed on to HadoopFileSystem
        """
        from pyarrow.fs import HadoopFileSystem

        fs = HadoopFileSystem(
            host=host,
            port=port,
            user=user,
            kerb_ticket=kerb_ticket,
            extra_conf=extra_conf,
        )
        super().__init__(fs=fs, **kwargs)

    @staticmethod
    def _get_kwargs_from_urls(path):
        ops = infer_storage_options(path)
        out = {}
        if ops.get("host", None):
            out["host"] = ops["host"]
        if ops.get("username", None):
            out["user"] = ops["username"]
        if ops.get("port", None):
            out["port"] = ops["port"]
        return out



import os

from spinsrv import spinfs

if "SPINPY_TEST_PUBLIC" not in os.environ:
    raise ValueError("must specify SPINPY_TEST_PUBLIC environment variable")
if "SPINPY_TEST_PRIVATE" not in os.environ:
    raise ValueError("must specify SPINPY_TEST_PRIVATE environment variable")
if "SPINPY_TEST_CITIZEN" not in os.environ:
    raise ValueError("must specify SPINPY_TEST_CITIZEN environment variable")
SPINPY_TEST_PUBLIC = os.environ["SPINPY_TEST_PUBLIC"]
SPINPY_TEST_PRIVATE = os.environ["SPINPY_TEST_PRIVATE"]
SPINPY_TEST_CITIZEN = os.environ["SPINPY_TEST_CITIZEN"]

print("THIS IS SMOKE TEST 4: IT TESTS A CHUNKED UPLOAD")

fs = spinfs.SpinFileSystem(
    spinfs.Config(
        public=SPINPY_TEST_PUBLIC,
        private=SPINPY_TEST_PRIVATE,
        citizen=SPINPY_TEST_CITIZEN,
    )
)
import shutil

src = open("test.mp4", "rb")
dst = fs.open("/test.mp4", "wb")
shutil.copyfileobj(src, dst)
src.close()
dst.close()



"""
    pygments.formatters.other
    ~~~~~~~~~~~~~~~~~~~~~~~~~

    Other formatters: NullFormatter, RawTokenFormatter.

    :copyright: Copyright 2006-2022 by the Pygments team, see AUTHORS.
    :license: BSD, see LICENSE for details.
"""

from pip._vendor.pygments.formatter import Formatter
from pip._vendor.pygments.util import get_choice_opt
from pip._vendor.pygments.token import Token
from pip._vendor.pygments.console import colorize

__all__ = ["NullFormatter", "RawTokenFormatter", "TestcaseFormatter"]


class NullFormatter(Formatter):
    """
    Output the text unchanged without any formatting.
    """

    name = "Text only"
    aliases = ["text", "null"]
    filenames = ["*.txt"]

    def format(self, tokensource, outfile):
        enc = self.encoding
        for ttype, value in tokensource:
            if enc:
                outfile.write(value.encode(enc))
            else:
                outfile.write(value)


class RawTokenFormatter(Formatter):
    r"""
    Format tokens as a raw representation for storing token streams.

    The format is ``tokentype<TAB>repr(tokenstring)\n``. The output can later
    be converted to a token stream with the `RawTokenLexer`, described in the
    :doc:`lexer list <lexers>`.

    Only two options are accepted:

    `compress`
        If set to ``'gz'`` or ``'bz2'``, compress the output with the given
        compression algorithm after encoding (default: ``''``).
    `error_color`
        If set to a color name, highlight error tokens using that color.  If
        set but with no value, defaults to ``'red'``.

        .. versionadded:: 0.11

    """
    name = "Raw tokens"
    aliases = ["raw", "tokens"]
    filenames = ["*.raw"]

    unicodeoutput = False

    def __init__(self, **options):
        Formatter.__init__(self, **options)
        # We ignore self.encoding if it is set, since it gets set for lexer
        # and formatter if given with -Oencoding on the command line.
        # The RawTokenFormatter outputs only ASCII. Override here.
        self.encoding = "ascii"  # let pygments.format() do the right thing
        self.compress = get_choice_opt(
            options, "compress", ["", "none", "gz", "bz2"], ""
        )
        self.error_color = options.get("error_color", None)
        if self.error_color is True:
            self.error_color = "red"
        if self.error_color is not None:
            try:
                colorize(self.error_color, "")
            except KeyError:
                raise ValueError("Invalid color %r specified" % self.error_color)

    def format(self, tokensource, outfile):
        try:
            outfile.write(b"")
        except TypeError:
            raise TypeError("The raw tokens formatter needs a binary " "output file")
        if self.compress == "gz":
            import gzip

            outfile = gzip.GzipFile("", "wb", 9, outfile)

            write = outfile.write
            flush = outfile.close
        elif self.compress == "bz2":
            import bz2

            compressor = bz2.BZ2Compressor(9)

            def write(text):
                outfile.write(compressor.compress(text))

            def flush():
                outfile.write(compressor.flush())
                outfile.flush()

        else:
            write = outfile.write
            flush = outfile.flush

        if self.error_color:
            for ttype, value in tokensource:
                line = b"%r\t%r\n" % (ttype, value)
                if ttype is Token.Error:
                    write(colorize(self.error_color, line))
                else:
                    write(line)
        else:
            for ttype, value in tokensource:
                write(b"%r\t%r\n" % (ttype, value))
        flush()


TESTCASE_BEFORE = """\
    def testNeedsName(lexer):
        fragment = %r
        tokens = [
"""
TESTCASE_AFTER = """\
        ]
        assert list(lexer.get_tokens(fragment)) == tokens
"""


class TestcaseFormatter(Formatter):
    """
    Format tokens as appropriate for a new testcase.

    .. versionadded:: 2.0
    """

    name = "Testcase"
    aliases = ["testcase"]

    def __init__(self, **options):
        Formatter.__init__(self, **options)
        if self.encoding is not None and self.encoding != "utf-8":
            raise ValueError("Only None and utf-8 are allowed encodings.")

    def format(self, tokensource, outfile):
        indentation = " " * 12
        rawbuf = []
        outbuf = []
        for ttype, value in tokensource:
            rawbuf.append(value)
            outbuf.append("%s(%s, %r),\n" % (indentation, ttype, value))

        before = TESTCASE_BEFORE % ("".join(rawbuf),)
        during = "".join(outbuf)
        after = TESTCASE_AFTER
        if self.encoding is None:
            outfile.write(before + during + after)
        else:
            outfile.write(before.encode("utf-8"))
            outfile.write(during.encode("utf-8"))
            outfile.write(after.encode("utf-8"))
        outfile.flush()



import os

from spinsrv import spinfs

if "SPINPY_TEST_PUBLIC" not in os.environ:
    raise ValueError("must specify SPINPY_TEST_PUBLIC environment variable")
if "SPINPY_TEST_PRIVATE" not in os.environ:
    raise ValueError("must specify SPINPY_TEST_PRIVATE environment variable")
if "SPINPY_TEST_CITIZEN" not in os.environ:
    raise ValueError("must specify SPINPY_TEST_CITIZEN environment variable")
SPINPY_TEST_PUBLIC = os.environ["SPINPY_TEST_PUBLIC"]
SPINPY_TEST_PRIVATE = os.environ["SPINPY_TEST_PRIVATE"]
SPINPY_TEST_CITIZEN = os.environ["SPINPY_TEST_CITIZEN"]

print(
    "THIS IS SMOKE TEST 3: IT TESTS (primarily file related) functionality of src/spin/spinfs.py"
)

fs = spinfs.SpinFileSystem(
    spinfs.Config(
        public=SPINPY_TEST_PUBLIC,
        private=SPINPY_TEST_PRIVATE,
        citizen=SPINPY_TEST_CITIZEN,
    )
)
f = fs.open("/notes.txt", "wb")
f.write(b"Hello world!")
f.close()
f = fs.open("/notes.txt", "rb")
b = f.read()
assert b == b"Hello world!", f"got {b}, wanted Hello world!"



import re
import sys
from dataclasses import dataclass
from functools import lru_cache
from typing import Iterator, List, Optional, Union

if sys.version_info >= (3, 8):
    from typing import Final
else:
    from typing_extensions import Final

from black.nodes import (
    CLOSING_BRACKETS,
    STANDALONE_COMMENT,
    WHITESPACE,
    container_of,
    first_leaf_of,
    preceding_leaf,
    syms,
)
from blib2to3.pgen2 import token
from blib2to3.pytree import Leaf, Node

# types
LN = Union[Leaf, Node]

FMT_OFF: Final = {"# fmt: off", "# fmt:off", "# yapf: disable"}
FMT_SKIP: Final = {"# fmt: skip", "# fmt:skip"}
FMT_PASS: Final = {*FMT_OFF, *FMT_SKIP}
FMT_ON: Final = {"# fmt: on", "# fmt:on", "# yapf: enable"}

COMMENT_EXCEPTIONS = {True: " !:#'", False: " !:#'%"}


@dataclass
class ProtoComment:
    """Describes a piece of syntax that is a comment.

    It's not a :class:`blib2to3.pytree.Leaf` so that:

    * it can be cached (`Leaf` objects should not be reused more than once as
      they store their lineno, column, prefix, and parent information);
    * `newlines` and `consumed` fields are kept separate from the `value`. This
      simplifies handling of special marker comments like ``# fmt: off/on``.
    """

    type: int  # token.COMMENT or STANDALONE_COMMENT
    value: str  # content of the comment
    newlines: int  # how many newlines before the comment
    consumed: int  # how many characters of the original leaf's prefix did we consume


def generate_comments(leaf: LN, *, preview: bool) -> Iterator[Leaf]:
    """Clean the prefix of the `leaf` and generate comments from it, if any.

    Comments in lib2to3 are shoved into the whitespace prefix.  This happens
    in `pgen2/driver.py:Driver.parse_tokens()`.  This was a brilliant implementation
    move because it does away with modifying the grammar to include all the
    possible places in which comments can be placed.

    The sad consequence for us though is that comments don't "belong" anywhere.
    This is why this function generates simple parentless Leaf objects for
    comments.  We simply don't know what the correct parent should be.

    No matter though, we can live without this.  We really only need to
    differentiate between inline and standalone comments.  The latter don't
    share the line with any code.

    Inline comments are emitted as regular token.COMMENT leaves.  Standalone
    are emitted with a fake STANDALONE_COMMENT token identifier.
    """
    for pc in list_comments(
        leaf.prefix, is_endmarker=leaf.type == token.ENDMARKER, preview=preview
    ):
        yield Leaf(pc.type, pc.value, prefix="\n" * pc.newlines)


@lru_cache(maxsize=4096)
def list_comments(
    prefix: str, *, is_endmarker: bool, preview: bool
) -> List[ProtoComment]:
    """Return a list of :class:`ProtoComment` objects parsed from the given `prefix`."""
    result: List[ProtoComment] = []
    if not prefix or "#" not in prefix:
        return result

    consumed = 0
    nlines = 0
    ignored_lines = 0
    for index, line in enumerate(re.split("\r?\n", prefix)):
        consumed += len(line) + 1  # adding the length of the split '\n'
        line = line.lstrip()
        if not line:
            nlines += 1
        if not line.startswith("#"):
            # Escaped newlines outside of a comment are not really newlines at
            # all. We treat a single-line comment following an escaped newline
            # as a simple trailing comment.
            if line.endswith("\\"):
                ignored_lines += 1
            continue

        if index == ignored_lines and not is_endmarker:
            comment_type = token.COMMENT  # simple trailing comment
        else:
            comment_type = STANDALONE_COMMENT
        comment = make_comment(line, preview=preview)
        result.append(
            ProtoComment(
                type=comment_type, value=comment, newlines=nlines, consumed=consumed
            )
        )
        nlines = 0
    return result


def make_comment(content: str, *, preview: bool) -> str:
    """Return a consistently formatted comment from the given `content` string.

    All comments (except for "##", "#!", "#:", '#'") should have a single
    space between the hash sign and the content.

    If `content` didn't start with a hash sign, one is provided.
    """
    content = content.rstrip()
    if not content:
        return "#"

    if content[0] == "#":
        content = content[1:]
    NON_BREAKING_SPACE = " "
    if (
        content
        and content[0] == NON_BREAKING_SPACE
        and not content.lstrip().startswith("type:")
    ):
        content = " " + content[1:]  # Replace NBSP by a simple space
    if content and content[0] not in COMMENT_EXCEPTIONS[preview]:
        content = " " + content
    return "#" + content


def normalize_fmt_off(node: Node, *, preview: bool) -> None:
    """Convert content between `# fmt: off`/`# fmt: on` into standalone comments."""
    try_again = True
    while try_again:
        try_again = convert_one_fmt_off_pair(node, preview=preview)


def convert_one_fmt_off_pair(node: Node, *, preview: bool) -> bool:
    """Convert content of a single `# fmt: off`/`# fmt: on` into a standalone comment.

    Returns True if a pair was converted.
    """
    for leaf in node.leaves():
        previous_consumed = 0
        for comment in list_comments(leaf.prefix, is_endmarker=False, preview=preview):
            if comment.value not in FMT_PASS:
                previous_consumed = comment.consumed
                continue
            # We only want standalone comments. If there's no previous leaf or
            # the previous leaf is indentation, it's a standalone comment in
            # disguise.
            if comment.value in FMT_PASS and comment.type != STANDALONE_COMMENT:
                prev = preceding_leaf(leaf)
                if prev:
                    if comment.value in FMT_OFF and prev.type not in WHITESPACE:
                        continue
                    if comment.value in FMT_SKIP and prev.type in WHITESPACE:
                        continue

            ignored_nodes = list(generate_ignored_nodes(leaf, comment, preview=preview))
            if not ignored_nodes:
                continue

            first = ignored_nodes[0]  # Can be a container node with the `leaf`.
            parent = first.parent
            prefix = first.prefix
            if comment.value in FMT_OFF:
                first.prefix = prefix[comment.consumed :]
            if comment.value in FMT_SKIP:
                first.prefix = ""
                standalone_comment_prefix = prefix
            else:
                standalone_comment_prefix = (
                    prefix[:previous_consumed] + "\n" * comment.newlines
                )
            hidden_value = "".join(str(n) for n in ignored_nodes)
            if comment.value in FMT_OFF:
                hidden_value = comment.value + "\n" + hidden_value
            if comment.value in FMT_SKIP:
                hidden_value += "  " + comment.value
            if hidden_value.endswith("\n"):
                # That happens when one of the `ignored_nodes` ended with a NEWLINE
                # leaf (possibly followed by a DEDENT).
                hidden_value = hidden_value[:-1]
            first_idx: Optional[int] = None
            for ignored in ignored_nodes:
                index = ignored.remove()
                if first_idx is None:
                    first_idx = index
            assert parent is not None, "INTERNAL ERROR: fmt: on/off handling (1)"
            assert first_idx is not None, "INTERNAL ERROR: fmt: on/off handling (2)"
            parent.insert_child(
                first_idx,
                Leaf(
                    STANDALONE_COMMENT,
                    hidden_value,
                    prefix=standalone_comment_prefix,
                ),
            )
            return True

    return False


def generate_ignored_nodes(
    leaf: Leaf, comment: ProtoComment, *, preview: bool
) -> Iterator[LN]:
    """Starting from the container of `leaf`, generate all leaves until `# fmt: on`.

    If comment is skip, returns leaf only.
    Stops at the end of the block.
    """
    if comment.value in FMT_SKIP:
        yield from _generate_ignored_nodes_from_fmt_skip(leaf, comment, preview=preview)
        return
    container: Optional[LN] = container_of(leaf)
    while container is not None and container.type != token.ENDMARKER:
        if is_fmt_on(container, preview=preview):
            return

        # fix for fmt: on in children
        if children_contains_fmt_on(container, preview=preview):
            for child in container.children:
                if isinstance(child, Leaf) and is_fmt_on(child, preview=preview):
                    if child.type in CLOSING_BRACKETS:
                        # This means `# fmt: on` is placed at a different bracket level
                        # than `# fmt: off`. This is an invalid use, but as a courtesy,
                        # we include this closing bracket in the ignored nodes.
                        # The alternative is to fail the formatting.
                        yield child
                    return
                if children_contains_fmt_on(child, preview=preview):
                    return
                yield child
        else:
            if container.type == token.DEDENT and container.next_sibling is None:
                # This can happen when there is no matching `# fmt: on` comment at the
                # same level as `# fmt: on`. We need to keep this DEDENT.
                return
            yield container
            container = container.next_sibling


def _generate_ignored_nodes_from_fmt_skip(
    leaf: Leaf, comment: ProtoComment, *, preview: bool
) -> Iterator[LN]:
    """Generate all leaves that should be ignored by the `# fmt: skip` from `leaf`."""
    prev_sibling = leaf.prev_sibling
    parent = leaf.parent
    # Need to properly format the leaf prefix to compare it to comment.value,
    # which is also formatted
    comments = list_comments(leaf.prefix, is_endmarker=False, preview=preview)
    if not comments or comment.value != comments[0].value:
        return
    if prev_sibling is not None:
        leaf.prefix = ""
        siblings = [prev_sibling]
        while "\n" not in prev_sibling.prefix and prev_sibling.prev_sibling is not None:
            prev_sibling = prev_sibling.prev_sibling
            siblings.insert(0, prev_sibling)
        yield from siblings
    elif (
        parent is not None and parent.type == syms.suite and leaf.type == token.NEWLINE
    ):
        # The `# fmt: skip` is on the colon line of the if/while/def/class/...
        # statements. The ignored nodes should be previous siblings of the
        # parent suite node.
        leaf.prefix = ""
        ignored_nodes: List[LN] = []
        parent_sibling = parent.prev_sibling
        while parent_sibling is not None and parent_sibling.type != syms.suite:
            ignored_nodes.insert(0, parent_sibling)
            parent_sibling = parent_sibling.prev_sibling
        # Special case for `async_stmt` where the ASYNC token is on the
        # grandparent node.
        grandparent = parent.parent
        if (
            grandparent is not None
            and grandparent.prev_sibling is not None
            and grandparent.prev_sibling.type == token.ASYNC
        ):
            ignored_nodes.insert(0, grandparent.prev_sibling)
        yield from iter(ignored_nodes)


def is_fmt_on(container: LN, preview: bool) -> bool:
    """Determine whether formatting is switched on within a container.
    Determined by whether the last `# fmt:` comment is `on` or `off`.
    """
    fmt_on = False
    for comment in list_comments(container.prefix, is_endmarker=False, preview=preview):
        if comment.value in FMT_ON:
            fmt_on = True
        elif comment.value in FMT_OFF:
            fmt_on = False
    return fmt_on


def children_contains_fmt_on(container: LN, *, preview: bool) -> bool:
    """Determine if children have formatting switched on."""
    for child in container.children:
        leaf = first_leaf_of(child)
        if leaf is not None and is_fmt_on(leaf, preview=preview):
            return True

    return False


def contains_pragma_comment(comment_list: List[Leaf]) -> bool:
    """
    Returns:
        True iff one of the comments in @comment_list is a pragma used by one
        of the more common static analysis tools for python (e.g. mypy, flake8,
        pylint).
    """
    for comment in comment_list:
        if comment.value.startswith(("# type:", "# noqa", "# pylint:")):
            return True

    return False



# -*- coding: utf-8 -*-
try:
    from ._version import version as __version__
except ImportError:
    __version__ = "unknown"

__all__ = ["easter", "parser", "relativedelta", "rrule", "tz", "utils", "zoneinfo"]



# SPDX-FileCopyrightText: 2015 Eric Larson
#
# SPDX-License-Identifier: Apache-2.0

from .file_cache import FileCache, SeparateBodyFileCache
from .redis_cache import RedisCache


__all__ = ["FileCache", "SeparateBodyFileCache", "RedisCache"]



import base64
import dataclasses
import datetime
from dateutil.parser import isoparse
import hashlib
import typing

# ZeroTime is the zero time used by default for spin types.
ZeroTime = datetime.datetime(1, 1, 1, 0, 0, tzinfo=datetime.timezone.utc)


# SHA256 creates a sum of the data
def SHA256(b: bytes):
    m = hashlib.sha256()
    m.update(b)
    return m.hexdigest()


""" CitizenName is a helper type for a spin CitizenName """
CitizenName = str
""" KeyType is a helper type for a spin KeyType """
KeyType = str
""" KeyName is a helper type for a spin KeyName """
KeyName = str


@dataclasses.dataclass
class Key:
    """
    Key is a helper class for a spin Key.
    """

    type: KeyType = ""
    citizen: CitizenName = ""
    name: KeyName = ""
    data: str = ""
    meta: str = ""
    created_at: datetime.datetime = ZeroTime
    expires_at: datetime.datetime = ZeroTime

    @staticmethod
    def from_json(j: dict):
        k = Key()
        k.unmarshal_json(j)
        return k

    def unmarshal_json(self, j: dict):
        if "Type" in j:
            self.type = KeyType(j["Type"])
        if "Citizen" in j:
            self.citizen = CitizenName(j["Citizen"])
        if "Name" in j:
            self.name = KeyName(j["Name"])
        if "Data" in j:
            self.data = j["Data"]
        if "Meta" in j:
            self.meta = j["Meta"]
        if "CreatedAt" in j:
            self.created_at = isoparse(j["CreatedAt"])
        if "ExpiresAt" in j:
            self.expires_at = isoparse(j["ExpiresAt"])

    def to_json(self):
        j = {}
        self.marshal_json(j)
        return j

    def marshal_json(self, j: dict):
        j["Type"] = self.type
        j["Citizen"] = self.citizen
        j["Name"] = self.name
        j["Data"] = self.data
        j["Meta"] = self.meta
        j["CreatedAt"] = self.created_at.isoformat()
        j["ExpiresAt"] = self.expires_at.isoformat()


@dataclasses.dataclass
class PrivateKey:
    key: Key = dataclasses.field(default_factory=Key)
    private: str = ""

    @staticmethod
    def from_json(j: dict):
        k = PrivateKey()
        k.unmarshal_json(j)
        return k

    def unmarshal_json(self, j: dict):
        self.key.unmarshal_json(j)
        if "Private" in j:
            self.private = j["Private"]

    def to_json(self):
        j = {}
        self.marshal_json(j)
        return j

    def marshal_json(self, j):
        j["Key"] = self.key.to_json()
        j["Private"] = self.private
        return j


@dataclasses.dataclass
class KeyWhichRequest:
    public: str = ""
    private: str = ""

    @staticmethod
    def from_json(j: dict):
        r = KeyWhichRequest()
        r.unmarshal_json(j)
        return r

    def unmarshal_json(self, j: dict):
        if "Public" in j:
            self.public = j["Public"]
        if "Private" in j:
            self.private = j["Private"]

    def to_json(self):
        j = {}
        self.marshal_json(j)
        return j

    def marshal_json(self, j):
        j["Public"] = self.public
        j["Private"] = self.private
        return j


@dataclasses.dataclass
class KeyWhichResponse:
    key: typing.Optional[Key] = None
    error: str = ""

    @staticmethod
    def from_json(j: dict):
        r = KeyWhichResponse()
        r.unmarshal_json(j)
        return r

    def unmarshal_json(self, j: dict):
        if "Key" in j and j["Key"] is not None:
            self.key = Key.from_json(j["Key"])
        if "Error" in j:
            self.error = j["Error"]

    def to_json(self):
        j = {}
        self.marshal_json(j)
        return j

    def marshal_json(self, j):
        j["Key"] = self.key.to_json()
        j["Private"] = self.private
        return j


@dataclasses.dataclass
class KeyTempRequest:
    public: str = ""
    private: str = ""
    duration: int = 0  # nanoseconds

    @staticmethod
    def from_json(j: dict):
        r = KeyTempRequest()
        r.unmarshal_json(j)
        return r

    def unmarshal_json(self, j: dict):
        if "Public" in j:
            self.public = j["Public"]
        if "Private" in j:
            self.private = j["Private"]
        if "Duration" in j:
            self.duration = j["Duration"]

    def to_json(self):
        j = {}
        self.marshal_json(j)
        return j

    def marshal_json(self, j):
        j["Public"] = self.public
        j["Private"] = self.private
        j["Duration"] = self.duration
        return j


@dataclasses.dataclass
class KeyTempResponse:
    key: typing.Optional[Key] = None
    private: str = ""
    error: str = ""

    @staticmethod
    def from_json(j: dict):
        r = KeyTempResponse()
        r.unmarshal_json(j)
        return r

    def unmarshal_json(self, j: dict):
        if "Key" in j and j["Key"] is not None:
            self.key = Key.from_json(j["Key"])
        if "Private" in j:
            self.private = j["Private"]
        if "Error" in j:
            self.error = j["Error"]

    def to_json(self):
        j = {}
        self.marshal_json(j)
        return j

    def marshal_json(self, j):
        j["Key"] = self.key.to_json()
        j["Private"] = self.private
        j["Duration"] = self.duration
        return j


Path = str
Ref = str
DirEntryType = str

EntryMissing: DirEntryType = ""
EntryDir: DirEntryType = "dir"
EntryFile: DirEntryType = "file"

SeqNotExist = -1
SeqIgnore = 0
SeqBase = 1
MaxBlockSize = 10 * (1024 * 1024)


@dataclasses.dataclass
class DirBlock:
    ref: Ref = ""
    offset: int = 0
    size: int = 0

    @staticmethod
    def from_json(j: dict):
        r = DirBlock()
        r.unmarshal_json(j)
        return r

    def unmarshal_json(self, j: dict):
        if "Ref" in j:
            self.ref = Ref(j["Ref"])
        if "Offset" in j:
            self.offset = j["Offset"]
        if "Size" in j:
            self.size = j["Size"]

    def to_json(self):
        j = {}
        self.marshal_json(j)
        return j

    def marshal_json(self, j):
        j["Ref"] = self.ref
        j["Offset"] = self.offset
        j["Size"] = self.size
        return j


@dataclasses.dataclass
class DirEntry:
    type: DirEntryType = EntryMissing
    blocks: list[DirBlock] = dataclasses.field(default_factory=list)
    citizen: str = ""
    path: Path = ""
    time: datetime.datetime = ZeroTime
    sequence: int = 0

    @staticmethod
    def from_json(j: dict):
        r = DirEntry()
        r.unmarshal_json(j)
        return r

    def unmarshal_json(self, j: dict):
        if "Type" in j:
            self.type = DirEntryType(j["Type"])
        if "Blocks" in j and j["Blocks"] is not None:
            self.blocks = [DirBlock.from_json(d) for d in j["Blocks"]]
        if "Citizen" in j:
            self.citizen = j["Citizen"]
        if "Path" in j:
            self.path = j["Path"]
        if "Time" in j:
            self.time = isoparse(j["Time"])
        if "Sequence" in j:
            self.sequence = j["Sequence"]

    def to_json(self):
        j = {}
        self.marshal_json(j)
        return j

    def marshal_json(self, j):
        j["Type"] = self.type
        j["Blocks"] = [b.to_json() for b in self.blocks]
        j["Citizen"] = self.citizen
        j["Path"] = self.path
        j["Time"] = self.time.isoformat()
        j["Sequence"] = self.sequence
        return j

    def size(self):
        return sum([b.size for b in self.blocks])


DirOperation = str

MissingDirOperation: DirOperation = ""
PutDirOperation: DirOperation = "put"
DelDirOperation: DirOperation = "del"


@dataclasses.dataclass
class DirOp:
    operation: DirOperation = MissingDirOperation
    dir_entry: DirEntry = dataclasses.field(default_factory=DirEntry)  # embedded

    @staticmethod
    def from_json(j: dict):
        r = DirOp()
        r.unmarshal_json(j)
        return r

    def unmarshal_json(self, j: dict):
        if "Operation" in j:
            self.operation = DirOperation(j["Operation"])
        # if "DirEntry" in j:
        #     self.dir_entry = DirEntry.from_json(j["DirEntry"])
        # else:
        self.dir_entry = DirEntry.from_json(j)

    def to_json(self):
        j = {}
        self.marshal_json(j)
        return j

    def marshal_json(self, j):
        j["Operation"] = self.operation
        self.dir_entry.marshal_json(j)
        return j


BitOperation = str

MissingBitOperation = ""
DelBitOperation = "del"
GetBitOperation = "get"
PutBitOperation = "put"


@dataclasses.dataclass
class BitOp:
    operation: BitOperation = MissingBitOperation
    ref: Ref = ""
    bytes: bytes = b""

    @staticmethod
    def from_json(j: dict):
        r = BitOp()
        r.unmarshal_json(j)
        return r

    def unmarshal_json(self, j: dict):
        if "Operation" in j:
            self.type = BitOperation(j["Operation"])
        if "Ref" in j:
            self.ref = Ref(j["Ref"])
        if "Bytes" in j and j["Bytes"] is not None:
            self.bytes = base64.b64decode(j["Bytes"])

    def to_json(self):
        j = {}
        self.marshal_json(j)
        return j

    def marshal_json(self, j):
        j["Operation"] = self.operation
        j["Ref"] = self.ref
        j["Bytes"] = base64.b64encode(self.bytes).decode()
        return j


@dataclasses.dataclass
class RefData:
    ref: Ref = ""
    volatile: bool = False
    duration: int = 0

    @staticmethod
    def from_json(j: dict):
        r = BitOp()
        r.unmarshal_json(j)
        return r

    def unmarshal_json(self, j: dict):
        if "Ref" in j:
            self.ref = Ref(j["Ref"])
        if "Volatile" in j:
            self.ref = Ref(j["Volatile"])
        if "Duration" in j:
            self.duration = j["Duration"]

    def to_json(self):
        j = {}
        self.marshal_json(j)
        return j

    def marshal_json(self, j):
        j["Ref"] = self.ref
        j["Volatile"] = self.volatile
        j["Duration"] = self.duration
        return j


@dataclasses.dataclass
class BitApplyRequest:
    public: str = ""
    private: str = ""
    ops: list[BitOp] = dataclasses.field(default_factory=list)

    @staticmethod
    def from_json(j: dict):
        r = BitApplyRequest()
        r.unmarshal_json(j)
        return r

    def unmarshal_json(self, j: dict):
        if "Public" in j:
            self.public = Ref(j["Public"])
        if "Private" in j:
            self.private = Ref(j["Private"])
        if "Ops" in j and j["Ops"] is not None:
            self.ops = [BitOp.from_json(o) for o in j["Ops"]]

    def to_json(self):
        j = {}
        self.marshal_json(j)
        return j

    def marshal_json(self, j):
        j["Public"] = self.public
        j["Private"] = self.private
        j["Ops"] = [o.to_json() for o in self.ops]
        return j


@dataclasses.dataclass
class BitApplyOutcome:
    ref_data: RefData = dataclasses.field(default_factory=RefData)
    bytes: bytes = b""
    Error: str = ""

    @staticmethod
    def from_json(j: dict):
        r = BitApplyOutcome()
        r.unmarshal_json(j)
        return r

    def unmarshal_json(self, j: dict):
        if "RefData" in j:
            self.ref_data = RefData.from_json(j["RefData"])
        if "Bytes" in j and j["Bytes"] is not None:
            self.bytes = base64.b64decode(j["Bytes"])
        if "Error" in j:
            self.error = j["Error"]

    def to_json(self):
        j = {}
        self.marshal_json(j)
        return j

    def marshal_json(self, j):
        self.ref_data.unmarshal_json(j)
        j["RefData"] = self.ref_data.to_json()
        j["Bytes"] = base64.b64encode(self.bytes).decode()
        j["Error"] = self.error
        return j


@dataclasses.dataclass
class BitApplyResponse:
    outcomes: list[BitApplyOutcome] = dataclasses.field(default_factory=list)
    error: str = ""

    @staticmethod
    def from_json(j: dict):
        r = BitApplyResponse()
        r.unmarshal_json(j)
        return r

    def unmarshal_json(self, j: dict):
        if "Outcomes" in j and j["Outcomes"] is not None:
            self.outcomes = [BitApplyOutcome.from_json(o) for o in j["Outcomes"]]
        if "Error" in j:
            self.error = j["Error"]

    def to_json(self):
        j = {}
        self.marshal_json(j)
        return j

    def marshal_json(self, j):
        j["Outcomes"] = [o.to_json() for o in self.outcomes]
        j["Error"] = self.error
        return j


@dataclasses.dataclass
class DirApplyRequest:
    public: str = ""
    private: str = ""
    ops: list[DirOp] = dataclasses.field(default_factory=list)

    @staticmethod
    def from_json(j: dict):
        r = DirApplyRequest()
        r.unmarshal_json(j)
        return r

    def unmarshal_json(self, j: dict):
        if "Public" in j:
            self.public = Ref(j["Public"])
        if "Private" in j:
            self.private = Ref(j["Private"])
        if "Ops" in j and j["Ops"] is not None:
            self.ops = [DirOp.from_json(o) for o in j["Ops"]]

    def to_json(self):
        j = {}
        self.marshal_json(j)
        return j

    def marshal_json(self, j):
        j["Public"] = self.public
        j["Private"] = self.private
        j["Ops"] = [o.to_json() for o in self.ops]
        return j


@dataclasses.dataclass
class DirApplyResponse:
    entries: list[DirEntry] = dataclasses.field(default_factory=list)
    error: str = ""

    @staticmethod
    def from_json(j: dict):
        r = DirApplyResponse()
        r.unmarshal_json(j)
        return r

    def unmarshal_json(self, j: dict):
        if "Entries" in j and j["Entries"] is not None:
            self.entries = [DirEntry.from_json(e) for e in j["Entries"]]
        if "Error" in j:
            self.error = j["Error"]

    def to_json(self):
        j = {}
        self.marshal_json(j)
        return j

    def marshal_json(self, j):
        j["Entries"] = [e.to_json() for e in self.entries]
        j["Error"] = self.error
        return j


@dataclasses.dataclass
class DirLookupRequest:
    public: str = ""
    private: str = ""
    citizen: CitizenName = ""
    path: Path = ""

    @staticmethod
    def from_json(j: dict):
        r = DirLookupRequest()
        r.unmarshal_json(j)
        return r

    def unmarshal_json(self, j: dict):
        if "Public" in j:
            self.public = Ref(j["Public"])
        if "Private" in j:
            self.private = Ref(j["Private"])
        if "Citizen" in j:
            self.citizen = CitizenName(j["CitizenName"])
        if "Path" in j:
            self.path = Path(j["Path"])

    def to_json(self):
        j = {}
        self.marshal_json(j)
        return j

    def marshal_json(self, j):
        j["Public"] = self.public
        j["Private"] = self.private
        j["Citizen"] = self.citizen
        j["Path"] = self.path
        return j


@dataclasses.dataclass
class DirLookupResponse:
    entry: typing.Optional[DirEntry] = None
    error: str = ""

    @staticmethod
    def from_json(j: dict):
        r = DirLookupResponse()
        r.unmarshal_json(j)
        return r

    def unmarshal_json(self, j: dict):
        if "Entry" in j and j["Entry"] is not None:
            self.entry = DirEntry.from_json(j["Entry"])
        if "Error" in j:
            self.error = j["Error"]

    def to_json(self):
        j = {}
        self.marshal_json(j)
        return j

    def marshal_json(self, j):
        j["Entry"] = self.entry.to_json()
        j["Error"] = self.error
        return j


@dataclasses.dataclass
class DirTreeRequest:
    public: str = ""
    private: str = ""
    citizen: CitizenName = ""
    path: Path = ""
    level: int = 0

    @staticmethod
    def from_json(j: dict):
        r = DirTreeRequest()
        r.unmarshal_json(j)
        return r

    def unmarshal_json(self, j: dict):
        if "Public" in j:
            self.public = Ref(j["Public"])
        if "Private" in j:
            self.private = Ref(j["Private"])
        if "Citizen" in j:
            self.citizen = CitizenName(j["CitizenName"])
        if "Path" in j:
            self.path = Path(j["Path"])
        if "Level" in j:
            self.level = j["Level"]

    def to_json(self):
        j = {}
        self.marshal_json(j)
        return j

    def marshal_json(self, j):
        j["Public"] = self.public
        j["Private"] = self.private
        j["Citizen"] = self.citizen
        j["Path"] = self.path
        j["Level"] = self.level
        return j


@dataclasses.dataclass
class DirTreeResponse:
    entries: list[DirEntry] = dataclasses.field(default_factory=list)
    error: str = ""

    @staticmethod
    def from_json(j: dict):
        r = DirTreeResponse()
        r.unmarshal_json(j)
        return r

    def unmarshal_json(self, j: dict):
        if "Entries" in j and j["Entries"] is not None:
            self.entries = [DirEntry.from_json(d) for d in j["Entries"]]
        if "Error" in j:
            self.error = j["Error"]

    def to_json(self):
        j = {}
        self.marshal_json(j)
        return j

    def marshal_json(self, j):
        j["Entries"] = [e.to_json() for e in self.entries]
        j["Error"] = self.error
        return j



"""
This module contains the project meta-data.
"""

__author__ = "Caleb P. Burns"
__copyright__ = "Copyright © 2013-2022 Caleb P. Burns"
__credits__ = [
    "dahlia <https://github.com/dahlia>",
    "highb <https://github.com/highb>",
    "029xue <https://github.com/029xue>",
    "mikexstudios <https://github.com/mikexstudios>",
    "nhumrich <https://github.com/nhumrich>",
    "davidfraser <https://github.com/davidfraser>",
    "demurgos <https://github.com/demurgos>",
    "ghickman <https://github.com/ghickman>",
    "nvie <https://github.com/nvie>",
    "adrienverge <https://github.com/adrienverge>",
    "AndersBlomdell <https://github.com/AndersBlomdell>",
    "thmxv <https://github.com/thmxv>",
    "wimglenn <https://github.com/wimglenn>",
    "hugovk <https://github.com/hugovk>",
    "dcecile <https://github.com/dcecile>",
    "mroutis <https://github.com/mroutis>",
    "jdufresne <https://github.com/jdufresne>",
    "groodt <https://github.com/groodt>",
    "ftrofin <https://github.com/ftrofin>",
    "pykong <https://github.com/pykong>",
    "nhhollander <https://github.com/nhhollander>",
    "KOLANICH <https://github.com/KOLANICH>",
    "JonjonHays <https://github.com/JonjonHays>",
    "Isaac0616 <https://github.com/Isaac0616>",
    "SebastiaanZ <https://github.com/SebastiaanZ>",
    "RoelAdriaans <https://github.com/RoelAdriaans>",
    "raviselker <https://github.com/raviselker>",
    "johanvergeer <https://github.com/johanvergeer>",
    "danjer <https://github.com/danjer>",
    "jhbuhrman <https://github.com/jhbuhrman>",
    "WPDOrdina <https://github.com/WPDOrdina>",
    "tirkarthi <https://github.com/tirkarthi>",
    "jayvdb <https://github.com/jayvdb>",
    "jwodder <https://github.com/jwodder>",
    "kloczek <https://github.com/kloczek>",
    "orens <https://github.com/orens>",
    "spMohanty <https://github.com/spMohanty>",
    "ichard26 <https://github.com/ichard26>",
    "jack1142 <https://github.com/jack1142>",
    "mgorny <https://github.com/mgorny>",
    "bzakdd <https://github.com/bzakdd>",
    "haimat <https://github.com/haimat>",
]
__license__ = "MPL 2.0"
__version__ = "0.10.2"



"""
This module provides the base definition for patterns.
"""

import dataclasses
import re
import warnings
from typing import (
    Any,
    AnyStr,
    Iterable,
    Iterator,
    Match as MatchHint,
    Optional,
    Pattern as PatternHint,
    Tuple,
    Union,
)


class Pattern(object):
    """
    The :class:`Pattern` class is the abstract definition of a pattern.
    """

    # Make the class dict-less.
    __slots__ = ("include",)

    def __init__(self, include: Optional[bool]) -> None:
        """
        Initializes the :class:`Pattern` instance.

        *include* (:class:`bool` or :data:`None`) is whether the matched
        files should be included (:data:`True`), excluded (:data:`False`),
        or is a null-operation (:data:`None`).
        """

        self.include = include
        """
		*include* (:class:`bool` or :data:`None`) is whether the matched
		files should be included (:data:`True`), excluded (:data:`False`),
		or is a null-operation (:data:`None`).
		"""

    def match(self, files: Iterable[str]) -> Iterator[str]:
        """
        DEPRECATED: This method is no longer used and has been replaced by
        :meth:`.match_file`. Use the :meth:`.match_file` method with a loop
        for similar results.

        Matches this pattern against the specified files.

        *files* (:class:`~collections.abc.Iterable` of :class:`str`)
        contains each file relative to the root directory (e.g.,
        :data:`"relative/path/to/file"`).

        Returns an :class:`~collections.abc.Iterable` yielding each matched
        file path (:class:`str`).
        """
        warnings.warn(
            (
                "{0.__module__}.{0.__qualname__}.match() is deprecated. Use "
                "{0.__module__}.{0.__qualname__}.match_file() with a loop for "
                "similar results."
            ).format(self.__class__),
            DeprecationWarning,
            stacklevel=2,
        )

        for file in files:
            if self.match_file(file) is not None:
                yield file

    def match_file(self, file: str) -> Optional[Any]:
        """
        Matches this pattern against the specified file.

        *file* (:class:`str`) is the normalized file path to match against.

        Returns the match result if *file* matched; otherwise, :data:`None`.
        """
        raise NotImplementedError(
            ("{0.__module__}.{0.__qualname__} must override match_file().").format(
                self.__class__
            )
        )


class RegexPattern(Pattern):
    """
    The :class:`RegexPattern` class is an implementation of a pattern
    using regular expressions.
    """

    # Keep the class dict-less.
    __slots__ = ("regex",)

    def __init__(
        self,
        pattern: Union[AnyStr, PatternHint],
        include: Optional[bool] = None,
    ) -> None:
        """
        Initializes the :class:`RegexPattern` instance.

        *pattern* (:class:`str`, :class:`bytes`, :class:`re.Pattern`, or
        :data:`None`) is the pattern to compile into a regular expression.

        *include* (:class:`bool` or :data:`None`) must be :data:`None`
        unless *pattern* is a precompiled regular expression (:class:`re.Pattern`)
        in which case it is whether matched files should be included
        (:data:`True`), excluded (:data:`False`), or is a null operation
        (:data:`None`).

                .. NOTE:: Subclasses do not need to support the *include*
                   parameter.
        """

        if isinstance(pattern, (str, bytes)):
            assert include is None, (
                "include:{!r} must be null when pattern:{!r} is a string."
            ).format(include, pattern)
            regex, include = self.pattern_to_regex(pattern)
            # NOTE: Make sure to allow a null regular expression to be
            # returned for a null-operation.
            if include is not None:
                regex = re.compile(regex)

        elif pattern is not None and hasattr(pattern, "match"):
            # Assume pattern is a precompiled regular expression.
            # - NOTE: Used specified *include*.
            regex = pattern

        elif pattern is None:
            # NOTE: Make sure to allow a null pattern to be passed for a
            # null-operation.
            assert include is None, (
                "include:{!r} must be null when pattern:{!r} is null."
            ).format(include, pattern)

        else:
            raise TypeError(
                "pattern:{!r} is not a string, re.Pattern, or None.".format(pattern)
            )

        super(RegexPattern, self).__init__(include)

        self.regex: PatternHint = regex
        """
		*regex* (:class:`re.Pattern`) is the regular expression for the
		pattern.
		"""

    def __eq__(self, other: "RegexPattern") -> bool:
        """
        Tests the equality of this regex pattern with *other* (:class:`RegexPattern`)
        by comparing their :attr:`~Pattern.include` and :attr:`~RegexPattern.regex`
        attributes.
        """
        if isinstance(other, RegexPattern):
            return self.include == other.include and self.regex == other.regex
        else:
            return NotImplemented

    def match_file(self, file: str) -> Optional["RegexMatchResult"]:
        """
        Matches this pattern against the specified file.

        *file* (:class:`str`)
        contains each file relative to the root directory (e.g., "relative/path/to/file").

        Returns the match result (:class:`RegexMatchResult`) if *file*
        matched; otherwise, :data:`None`.
        """
        if self.include is not None:
            match = self.regex.match(file)
            if match is not None:
                return RegexMatchResult(match)

        return None

    @classmethod
    def pattern_to_regex(cls, pattern: str) -> Tuple[str, bool]:
        """
        Convert the pattern into an uncompiled regular expression.

        *pattern* (:class:`str`) is the pattern to convert into a regular
        expression.

        Returns the uncompiled regular expression (:class:`str` or :data:`None`),
        and whether matched files should be included (:data:`True`),
        excluded (:data:`False`), or is a null-operation (:data:`None`).

                .. NOTE:: The default implementation simply returns *pattern* and
                   :data:`True`.
        """
        return pattern, True


@dataclasses.dataclass()
class RegexMatchResult(object):
    """
    The :class:`RegexMatchResult` data class is used to return information
    about the matched regular expression.
    """

    # Keep the class dict-less.
    __slots__ = ("match",)

    match: MatchHint
    """
	*match* (:class:`re.Match`) is the regex match result.
	"""



from .core import encode, decode, alabel, ulabel, IDNAError
import codecs
import re
from typing import Tuple, Optional

_unicode_dots_re = re.compile("[\u002e\u3002\uff0e\uff61]")


class Codec(codecs.Codec):
    def encode(self, data: str, errors: str = "strict") -> Tuple[bytes, int]:
        if errors != "strict":
            raise IDNAError('Unsupported error handling "{}"'.format(errors))

        if not data:
            return b"", 0

        return encode(data), len(data)

    def decode(self, data: bytes, errors: str = "strict") -> Tuple[str, int]:
        if errors != "strict":
            raise IDNAError('Unsupported error handling "{}"'.format(errors))

        if not data:
            return "", 0

        return decode(data), len(data)


class IncrementalEncoder(codecs.BufferedIncrementalEncoder):
    def _buffer_encode(self, data: str, errors: str, final: bool) -> Tuple[str, int]:  # type: ignore
        if errors != "strict":
            raise IDNAError('Unsupported error handling "{}"'.format(errors))

        if not data:
            return "", 0

        labels = _unicode_dots_re.split(data)
        trailing_dot = ""
        if labels:
            if not labels[-1]:
                trailing_dot = "."
                del labels[-1]
            elif not final:
                # Keep potentially unfinished label until the next call
                del labels[-1]
                if labels:
                    trailing_dot = "."

        result = []
        size = 0
        for label in labels:
            result.append(alabel(label))
            if size:
                size += 1
            size += len(label)

        # Join with U+002E
        result_str = ".".join(result) + trailing_dot  # type: ignore
        size += len(trailing_dot)
        return result_str, size


class IncrementalDecoder(codecs.BufferedIncrementalDecoder):
    def _buffer_decode(self, data: str, errors: str, final: bool) -> Tuple[str, int]:  # type: ignore
        if errors != "strict":
            raise IDNAError('Unsupported error handling "{}"'.format(errors))

        if not data:
            return ("", 0)

        labels = _unicode_dots_re.split(data)
        trailing_dot = ""
        if labels:
            if not labels[-1]:
                trailing_dot = "."
                del labels[-1]
            elif not final:
                # Keep potentially unfinished label until the next call
                del labels[-1]
                if labels:
                    trailing_dot = "."

        result = []
        size = 0
        for label in labels:
            result.append(ulabel(label))
            if size:
                size += 1
            size += len(label)

        result_str = ".".join(result) + trailing_dot
        size += len(trailing_dot)
        return (result_str, size)


class StreamWriter(Codec, codecs.StreamWriter):
    pass


class StreamReader(Codec, codecs.StreamReader):
    pass


def getregentry() -> codecs.CodecInfo:
    # Compatibility as a search_function for codecs.register()
    return codecs.CodecInfo(
        name="idna",
        encode=Codec().encode,  # type: ignore
        decode=Codec().decode,  # type: ignore
        incrementalencoder=IncrementalEncoder,
        incrementaldecoder=IncrementalDecoder,
        streamwriter=StreamWriter,
        streamreader=StreamReader,
    )



"""
    pygments.formatters.groff
    ~~~~~~~~~~~~~~~~~~~~~~~~~

    Formatter for groff output.

    :copyright: Copyright 2006-2022 by the Pygments team, see AUTHORS.
    :license: BSD, see LICENSE for details.
"""

import math
from pip._vendor.pygments.formatter import Formatter
from pip._vendor.pygments.util import get_bool_opt, get_int_opt

__all__ = ["GroffFormatter"]


class GroffFormatter(Formatter):
    """
    Format tokens with groff escapes to change their color and font style.

    .. versionadded:: 2.11

    Additional options accepted:

    `style`
        The style to use, can be a string or a Style subclass (default:
        ``'default'``).

    `monospaced`
        If set to true, monospace font will be used (default: ``true``).

    `linenos`
        If set to true, print the line numbers (default: ``false``).

    `wrap`
        Wrap lines to the specified number of characters. Disabled if set to 0
        (default: ``0``).
    """

    name = "groff"
    aliases = ["groff", "troff", "roff"]
    filenames = []

    def __init__(self, **options):
        Formatter.__init__(self, **options)

        self.monospaced = get_bool_opt(options, "monospaced", True)
        self.linenos = get_bool_opt(options, "linenos", False)
        self._lineno = 0
        self.wrap = get_int_opt(options, "wrap", 0)
        self._linelen = 0

        self.styles = {}
        self._make_styles()

    def _make_styles(self):
        regular = "\\f[CR]" if self.monospaced else "\\f[R]"
        bold = "\\f[CB]" if self.monospaced else "\\f[B]"
        italic = "\\f[CI]" if self.monospaced else "\\f[I]"

        for ttype, ndef in self.style:
            start = end = ""
            if ndef["color"]:
                start += "\\m[%s]" % ndef["color"]
                end = "\\m[]" + end
            if ndef["bold"]:
                start += bold
                end = regular + end
            if ndef["italic"]:
                start += italic
                end = regular + end
            if ndef["bgcolor"]:
                start += "\\M[%s]" % ndef["bgcolor"]
                end = "\\M[]" + end

            self.styles[ttype] = start, end

    def _define_colors(self, outfile):
        colors = set()
        for _, ndef in self.style:
            if ndef["color"] is not None:
                colors.add(ndef["color"])

        for color in colors:
            outfile.write(".defcolor " + color + " rgb #" + color + "\n")

    def _write_lineno(self, outfile):
        self._lineno += 1
        outfile.write("%s% 4d " % (self._lineno != 1 and "\n" or "", self._lineno))

    def _wrap_line(self, line):
        length = len(line.rstrip("\n"))
        space = "     " if self.linenos else ""
        newline = ""

        if length > self.wrap:
            for i in range(0, math.floor(length / self.wrap)):
                chunk = line[i * self.wrap : i * self.wrap + self.wrap]
                newline += chunk + "\n" + space
            remainder = length % self.wrap
            if remainder > 0:
                newline += line[-remainder - 1 :]
                self._linelen = remainder
        elif self._linelen + length > self.wrap:
            newline = ("\n" + space) + line
            self._linelen = length
        else:
            newline = line
            self._linelen += length

        return newline

    def _escape_chars(self, text):
        text = (
            text.replace("\\", "\\[u005C]")
            .replace(".", "\\[char46]")
            .replace("'", "\\[u0027]")
            .replace("`", "\\[u0060]")
            .replace("~", "\\[u007E]")
        )
        copy = text

        for char in copy:
            if len(char) != len(char.encode()):
                uni = (
                    char.encode("unicode_escape")
                    .decode()[1:]
                    .replace("x", "u00")
                    .upper()
                )
                text = text.replace(char, "\\[u" + uni[1:] + "]")

        return text

    def format_unencoded(self, tokensource, outfile):
        self._define_colors(outfile)

        outfile.write(".nf\n\\f[CR]\n")

        if self.linenos:
            self._write_lineno(outfile)

        for ttype, value in tokensource:
            while ttype not in self.styles:
                ttype = ttype.parent
            start, end = self.styles[ttype]

            for line in value.splitlines(True):
                if self.wrap > 0:
                    line = self._wrap_line(line)

                if start and end:
                    text = self._escape_chars(line.rstrip("\n"))
                    if text != "":
                        outfile.write("".join((start, text, end)))
                else:
                    outfile.write(self._escape_chars(line.rstrip("\n")))

                if line.endswith("\n"):
                    if self.linenos:
                        self._write_lineno(outfile)
                        self._linelen = 0
                    else:
                        outfile.write("\n")
                        self._linelen = 0

        outfile.write("\n.fi")
